from logging import Logger

from cfgtools import Group
from modelcommon.model import SitefResponse

from dialogmanager import DialogManager
from mbcontextmessagehandler import MbContextMessageBus
from typing import (
    Any,
    Union,
    Optional,
)

from model import (
    CancelPaymentOrderData,
    ConfirmFinishPaymentData,
    TestConnectionData,
    StartPaymentData,
)
from paymentprocessor import PaymentProcessor
from util import Cache
from ._SitefPaymentProcessor import SitefPaymentProcessor


class SitefDigitalWalletsPaymentProcessor(PaymentProcessor):
    def __init__(
        self,
        message_bus,        # type: MbContextMessageBus
        dialog_manager,     # type: DialogManager
        config,             # type: Group
        logger,             # type: Logger
        cache,              # type: Cache
    ):
        # type: (...) -> None

        super(SitefDigitalWalletsPaymentProcessor, self).__init__(message_bus, dialog_manager, config, logger, cache)
        self.sitef_processor = SitefPaymentProcessor(message_bus, dialog_manager, config, logger, cache)
        self.show_qr_code = config.find_value("ShowQRCode", "false").lower() == "true"
        self.allowed_payments = config.find_values("AllowedPayments") or []
        self.forbidden_payments = config.find_values("ForbiddenPayments") or []

        default_disabled_payments = ['10', '17', '18', '19', '27', '28', '29', '34', '35', '36', '42']
        self.disabled_payments = config.find_values("DisabledPayments") or default_disabled_payments

        self.additional_data = self._get_additional_data()

    def connection_status(
        self,
        test_connection_data    # type: TestConnectionData
    ):
        # type: (...) -> bool

        return self.sitef_processor.connection_status(test_connection_data=test_connection_data)

    def add_payment(
        self,                   # type: SitefDigitalWalletsPaymentProcessor
        start_payment_data,     # type: StartPaymentData
    ):
        # type: (...) -> Any[Union[SitefResponse, str]]

        return self.sitef_processor.add_payment(
            start_payment_data=start_payment_data,
            additional_params=self.additional_data,
        )

    def cancel_payment_order(
        self,
        cancel_payment_order_data,  # type: CancelPaymentOrderData
        additional_params="",       # type: Optional[str]
    ):
        # type: (...) -> Any[Union[SitefResponse, str]]

        return self.sitef_processor.cancel_payment_order(cancel_payment_order_data=cancel_payment_order_data)

    def cancel_payment(self):
        return self.sitef_processor.cancel_payment()

    def finish_payment(self, finish_payment_data):
        return self.sitef_processor.finish_payment(finish_payment_data)

    def confirm_finish_payment(
        self,
        confirm_finish_payment_data     # type: ConfirmFinishPaymentData
    ):
        # type: (...) -> None

        pass

    def get_name(self):
        return self.sitef_processor.get_name()

    def get_document(
        self,
        pos_id      # type: str
    ):
        # type: (...) -> str

        return self.sitef_processor.get_document(pos_id)

    def stop(self):
        self.sitef_processor.stop()

    def get_last_receipt(
        self,
        eft_processor,     # type: str
    ):
        # type: (...) -> Optional[str]

        return self.sitef_processor.get_last_receipt(eft_processor=eft_processor)

    def _get_additional_data(self):
        additional_data_list = []

        if self.disabled_payments:
            additional_data_list.append("[{}]".format(';'.join(self.disabled_payments)))

        if self.show_qr_code:
            additional_data_list.append("{DevolveStringQRCode=1}")

        if self.allowed_payments:
            additional_data_list.append("{{CarteirasDigitaisHabilitadas={}}}".format(';'.join(self.allowed_payments)))

        if self.forbidden_payments:
            additional_data_list.append("{{RestricoesCarteirasDigitais={}}}".format(';'.join(self.forbidden_payments)))

        return ';'.join(additional_data_list)

    def get_qr_code_config(
        self,   # type: SitefDigitalWalletsPaymentProcessor
    ):
        # type: () -> bool

        return self.sitef_processor.get_qr_code_config()




class SitefResponse(object):
    def __init__(self):
        # type: (...) -> None

        self.nsu = None
        self.nsu_host = None
        self.nsu_origin = None
        self.transaction_value = None
        self.customer_receipt = None
        self.merchant_receipt = None
        self.transaction_identifier = None




# Copyright (C) Dnspython Contributors, see LICENSE for text of ISC license

# Copyright (C) 2003-2017 Nominum, Inc.
#
# Permission to use, copy, modify, and distribute this software and its
# documentation for any purpose with or without fee is hereby granted,
# provided that the above copyright notice and this permission notice
# appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND NOMINUM DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL NOMINUM BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT
# OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

"""Generic Internet address helper functions."""

import socket

import dns.ipv4
import dns.ipv6

from ._compat import maybe_ord

# We assume that AF_INET is always defined.

AF_INET = socket.AF_INET

# AF_INET6 might not be defined in the socket module, but we need it.
# We'll try to use the socket module's value, and if it doesn't work,
# we'll use our own value.

try:
    AF_INET6 = socket.AF_INET6
except AttributeError:
    AF_INET6 = 9999


def inet_pton(family, text):
    """Convert the textual form of a network address into its binary form.

    *family* is an ``int``, the address family.

    *text* is a ``text``, the textual address.

    Raises ``NotImplementedError`` if the address family specified is not
    implemented.

    Returns a ``binary``.
    """

    if family == AF_INET:
        return dns.ipv4.inet_aton(text)
    elif family == AF_INET6:
        return dns.ipv6.inet_aton(text)
    else:
        raise NotImplementedError


def inet_ntop(family, address):
    """Convert the binary form of a network address into its textual form.

    *family* is an ``int``, the address family.

    *address* is a ``binary``, the network address in binary form.

    Raises ``NotImplementedError`` if the address family specified is not
    implemented.

    Returns a ``text``.
    """

    if family == AF_INET:
        return dns.ipv4.inet_ntoa(address)
    elif family == AF_INET6:
        return dns.ipv6.inet_ntoa(address)
    else:
        raise NotImplementedError


def af_for_address(text):
    """Determine the address family of a textual-form network address.

    *text*, a ``text``, the textual address.

    Raises ``ValueError`` if the address family cannot be determined
    from the input.

    Returns an ``int``.
    """

    try:
        dns.ipv4.inet_aton(text)
        return AF_INET
    except Exception:
        try:
            dns.ipv6.inet_aton(text)
            return AF_INET6
        except:
            raise ValueError


def is_multicast(text):
    """Is the textual-form network address a multicast address?

    *text*, a ``text``, the textual address.

    Raises ``ValueError`` if the address family cannot be determined
    from the input.

    Returns a ``bool``.
    """

    try:
        first = maybe_ord(dns.ipv4.inet_aton(text)[0])
        return first >= 224 and first <= 239
    except Exception:
        try:
            first = maybe_ord(dns.ipv6.inet_aton(text)[0])
            return first == 255
        except Exception:
            raise ValueError



# -*- coding: utf-8 -*-
"""
This module offers general convenience and utility functions for dealing with
datetimes.

.. versionadded:: 2.7.0
"""
from __future__ import unicode_literals

from datetime import datetime, time


def today(tzinfo=None):
    """
    Returns a :py:class:`datetime` representing the current day at midnight

    :param tzinfo:
        The time zone to attach (also used to determine the current day).

    :return:
        A :py:class:`datetime.datetime` object representing the current day
        at midnight.
    """

    dt = datetime.now(tzinfo)
    return datetime.combine(dt.date(), time(0, tzinfo=tzinfo))


def default_tzinfo(dt, tzinfo):
    """
    Sets the the ``tzinfo`` parameter on naive datetimes only

    This is useful for example when you are provided a datetime that may have
    either an implicit or explicit time zone, such as when parsing a time zone
    string.

    .. doctest::

        >>> from dateutil.tz import tzoffset
        >>> from dateutil.parser import parse
        >>> from dateutil.utils import default_tzinfo
        >>> dflt_tz = tzoffset("EST", -18000)
        >>> print(default_tzinfo(parse('2014-01-01 12:30 UTC'), dflt_tz))
        2014-01-01 12:30:00+00:00
        >>> print(default_tzinfo(parse('2014-01-01 12:30'), dflt_tz))
        2014-01-01 12:30:00-05:00

    :param dt:
        The datetime on which to replace the time zone

    :param tzinfo:
        The :py:class:`datetime.tzinfo` subclass instance to assign to
        ``dt`` if (and only if) it is naive.

    :return:
        Returns an aware :py:class:`datetime.datetime`.
    """
    if dt.tzinfo is not None:
        return dt
    else:
        return dt.replace(tzinfo=tzinfo)


def within_delta(dt1, dt2, delta):
    """
    Useful for comparing two datetimes that may a negilible difference
    to be considered equal.
    """
    delta = abs(delta)
    difference = dt1 - dt2
    return -delta <= difference <= delta



from production.box._OrderChangerProductionBox import OrderChangerProductionBox

from ._AllowedForbiddenFilter import AllowedForbiddenFilter


class TagFilterBox(OrderChangerProductionBox):
    def __init__(
        self,
        name,
        sons,
        allowed_tags,
        forbidden_tags,
        ignore_tags,
        allow_combos,
        keep_item_with_sons,
        keep_combos_without_sons,
        inherit_comments,
        logger=None
    ):
        super(TagFilterBox, self).__init__(name, sons, logger)
        self.filter = AllowedForbiddenFilter(
            allowed_values=allowed_tags,
            forbidden_values=forbidden_tags,
            ignore_values=ignore_tags,
            keep_son_parts=allow_combos,
            keep_item_with_sons=keep_item_with_sons,
            inherit_comments=inherit_comments,
        )

        self.keep_combos_without_sons = keep_combos_without_sons.lower() == 'true'

    def change_order(self, order):
        combo_without_sons = []
        for item in order.items:
            if item.is_combo() and not item.has_some_product_child():
                combo_without_sons.append(item.get_line_id())

        filtered_order = self.filter.filter_items(order, lambda x: x.get_tags())

        filtered_combos_order = []
        for item in filtered_order.items:
            if not item.is_combo() or self.keep_combos_without_sons:
                filtered_combos_order.append(item)
                continue

            if item.has_some_product_child() or item.get_line_id() in combo_without_sons:
                filtered_combos_order.append(item)

        filtered_order.items = filtered_combos_order
        return filtered_order



# uen.py - functions for handling Singapore UEN numbers
# coding: utf-8
#
# Copyright (C) 2020 Leandro Regueiro
# Copyright (C) 2020 Arthur de Jong
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

"""UEN (Singapore's Unique Entity Number).

The Unique Entity Number (UEN) is a 9 or 10 digit identification issued by
the government of Singapore to businesses that operate within Singapore.


Accounting and Corporate Regulatory Authority (ACRA)

There are three different formats:

* Business (ROB): It consists of 8 digits followed by a check letter.
* Local Company (ROC): It consists of 9 digits (the 4 leftmost digits
  represent the year of issuance) followed by a check letter.
* Others: Consists of 10 characters, begins with either the R letter, or the
  S letter or the T letter followed by 2 digits representing the last two
  digits of the issuance year, followed by two letters representing the
  entity type, 4 digits and finally a check letter.

More information:

* https://www.oecd.org/tax/automatic-exchange/crs-implementation-and-assistance/tax-identification-numbers/Singapore-TIN.pdf
* https://www.uen.gov.sg/ueninternet/faces/pages/admin/aboutUEN.jspx

>>> validate('00192200M')
'00192200M'
>>> validate('197401143C')
'197401143C'
>>> validate('S16FC0121D')
'S16FC0121D'
>>> validate('T01FC6132D')
'T01FC6132D'
>>> validate('123456')
Traceback (most recent call last):
    ...
InvalidLength: ...
"""  # noqa: E501

# There are some references to special 10-digit (or 7-digit) numbers that
# start with an F for foreign companies but it is unclear whether this is
# still current and not even examples of these numbers could be found.

from datetime import datetime

from stdnum.exceptions import *
from stdnum.util import clean, isdigits


OTHER_UEN_ENTITY_TYPES = (
    'CC', 'CD', 'CH', 'CL', 'CM', 'CP', 'CS', 'CX', 'DP', 'FB', 'FC', 'FM',
    'FN', 'GA', 'GB', 'GS', 'HS', 'LL', 'LP', 'MB', 'MC', 'MD', 'MH', 'MM',
    'MQ', 'NB', 'NR', 'PA', 'PB', 'PF', 'RF', 'RP', 'SM', 'SS', 'TC', 'TU',
    'VH', 'XL',
)


def compact(number):
    """Convert the number to the minimal representation.

    This converts to uppercase and removes surrounding whitespace. It
    also replaces the whitespace in UEN for foreign companies with
    zeroes.
    """
    return clean(number).upper().strip()


def calc_business_check_digit(number):
    """Calculate the check digit for the Business (ROB) number."""
    number = compact(number)
    weights = (10, 4, 9, 3, 8, 2, 7, 1)
    return 'XMKECAWLJDB'[sum(int(n) * w for n, w in zip(number, weights)) % 11]


def _validate_business(number):
    """Perform validation on UEN - Business (ROB) numbers."""
    if not isdigits(number[:-1]):
        raise InvalidFormat()
    if not number[-1].isalpha():
        raise InvalidFormat()
    if number[-1] != calc_business_check_digit(number):
        raise InvalidChecksum()
    return number


def calc_local_company_check_digit(number):
    """Calculate the check digit for the Local Company (ROC) number."""
    number = compact(number)
    weights = (10, 8, 6, 4, 9, 7, 5, 3, 1)
    return 'ZKCMDNERGWH'[sum(int(n) * w for n, w in zip(number, weights)) % 11]


def _validate_local_company(number):
    """Perform validation on UEN - Local Company (ROC) numbers."""
    if not isdigits(number[:-1]):
        raise InvalidFormat()
    current_year = str(datetime.now().year)
    if number[:4] > current_year:
        raise InvalidComponent()
    if number[-1] != calc_local_company_check_digit(number):
        raise InvalidChecksum()
    return number


def calc_other_check_digit(number):
    """Calculate the check digit for the other entities number."""
    number = compact(number)
    alphabet = 'ABCDEFGHJKLMNPQRSTUVWX0123456789'
    weights = (4, 3, 5, 3, 10, 2, 2, 5, 7)
    return alphabet[(sum(alphabet.index(n) * w for n, w in zip(number, weights)) - 5) % 11]


def _validate_other(number):
    """Perform validation on other UEN numbers."""
    if number[0] not in ('R', 'S', 'T'):
        raise InvalidComponent()
    if not isdigits(number[1:3]):
        raise InvalidFormat()
    current_year = str(datetime.now().year)
    if number[0] == 'T' and number[1:3] > current_year[2:]:
        raise InvalidComponent()
    if number[3:5] not in OTHER_UEN_ENTITY_TYPES:
        raise InvalidComponent()
    if not isdigits(number[5:-1]):
        raise InvalidFormat()
    if number[-1] != calc_other_check_digit(number):
        raise InvalidChecksum()
    return number


def validate(number):
    """Check if the number is a valid Singapore UEN number."""
    number = compact(number)
    if len(number) not in (9, 10):
        raise InvalidLength()
    if len(number) == 9:
        return _validate_business(number)
    if isdigits(number[0]):
        return _validate_local_company(number)
    return _validate_other(number)


def is_valid(number):
    """Check if the number is a valid Singapore UEN number."""
    try:
        return bool(validate(number))
    except ValidationError:
        return False


def format(number):
    """Reformat the number to the standard presentation format."""
    return compact(number)



from ._AlignCommandProcessor import AlignCommandProcessor
from ._BoldCommandProcessor import BoldCommandProcessor
from ._RepeatCommandProcessor import RepeatCommandProcessor
from ._FontCommandProcessor import FontCommandProcessor
from ._NewLineCommandProcessor import NewLineCommandProcessor
from ._CutPaperCommandProcessor import CutPaperCommandProcessor



import itertools

from mbcontextmessagehandler import MbContextMessageBus
from messagebus import Event
from model.event_dispatch import EventDispatch
from model.services import EventDispatcherService

EVENT_XML_TEMPLATE = ''.join(map(str.strip,
                                 ('<Event subject="AUDITLOGGER" type="{event_type}">',
                                  '    <AuditLog rowId="{row_id}" period="{business_period}" encoded="{encoded:d}" orderId="{order_id}" createdAt="{created_at_gmt}" evtTime="{event_time}">',
                                  '          {event_data}',
                                  '    </AuditLog>',
                                  '</Event>')))


class EventDispatcher(EventDispatcherService):
    def __init__(self, message_bus):
        # type: (MbContextMessageBus) -> None
        self.message_bus = message_bus

    def dispatch_event(self, event_to_dispatch):
        # type: (EventDispatch) -> None

        params = dict(itertools.chain.from_iterable((
            event_to_dispatch.__dict__.iteritems(),
            event_to_dispatch.log_entry.__dict__.iteritems())
        ))
        formatted_xml = EVENT_XML_TEMPLATE.format(**params)

        self.message_bus.publish_event(Event("AUDITLOGGER", event_to_dispatch.log_entry.event_type, formatted_xml))



# -*- coding: utf-8 -*-

import sysactions
from systools import sys_log_exception


@sysactions.action
def do_void_kiosk_order(pos_id, show_message="false"):
    try:
        model = sysactions.get_model(pos_id)
        pos_ot = sysactions.get_posot(model)

        if sysactions.has_current_order(model):
            order = sysactions.get_current_order(model)
            if order.get("state") not in ["IN_PROGRESS", "TOTALED"]:
                return

            show_message = (show_message or "false").lower() == "true"
            if show_message and float(order.get("totalGross")) > 0:
                ret = sysactions.show_messagebox(pos_id, "$KIOSK_CANCEL_QUESTION", "$VOID_SALE", "$YES|$NO")
                if ret != 0:
                    raise sysactions.StopAction()

            pos_ot.voidOrder(pos_id)

        return True
    except BaseException as _:
        sys_log_exception("Error voiding kiosk order")
        return False



from helper import ExtendedEnum
from messagebus import (
    TokenCreator,
    TokenPriority,
)

MESSAGE_GROUP = "114"


class ListenedTokens(ExtendedEnum):
    TK_ALERTS_GET_CONFIGURED_ALERTS = TokenCreator.create_token(TokenPriority.low, MESSAGE_GROUP, "1")
    TK_ALERTS_UPDATE_ALERTS = TokenCreator.create_token(TokenPriority.low, MESSAGE_GROUP, "2")



# -*- coding: utf-8 -*-

from datetime import datetime

from application.model import UpdateType, UpdateStatus
from application.model.configuration import Configurations
from application.model.customexception import ErrorInsertingNewUpdateVersion, ErrorRestartingHV, ErrorGettingUsersList,\
    ErrorUpdatingUserControlType
from mbcontextmessagehandler import MbContextMessageBus
from msgbus import (
    TK_SYS_ACK,
    TK_POS_LISTUSERS,
    TK_POS_GETMODEL
)
from messagebus import Message, DataType
from mwhelper import (
    BaseRepository,
    restart_system,
)
from persistence import Connection
from typing import (
    Any,
    Optional,
    Union
)
from xml.etree import cElementTree as etree

from ._CommonPOSRepository import CommonPOSRepository


class LoaderUpdaterPOSRepository(BaseRepository):
    def __init__(self, message_bus, configs):
        # type: (MbContextMessageBus, Configurations) -> None

        super(LoaderUpdaterPOSRepository, self).__init__(message_bus.mbcontext)

        self.message_bus = message_bus
        self.configs = configs
        self.update_type = UpdateType.loader
        self.update_value = self.update_type.value
        self.local_configs = configs.updaters[self.update_type.name]
        self.common_pos_repository = CommonPOSRepository(message_bus.mbcontext)

    def insert_new_loader_update(self, update_id, update_name):
        # type: (str, str) -> None

        try:
            def inner_func(conn):
                # type: (Connection) -> None

                date_now = str(datetime.now().isoformat())
                query = """
                INSERT INTO UpdatesController
                (UpdateId, UpdateName, TypeId, StatusId, ObtainedDate, DownloadedDate, BackupDate)
                VALUES ('{0}', '{1}', '{2}', '{3}', '{4}', '{4}', '{4}')"""
                query_format = query.format(update_id, update_name, self.update_value, UpdateStatus.pending,
                                            date_now)
                conn.query(query_format)

            self.execute_with_connection(inner_func, service_name=self.configs.persistence_name)
        except Exception as ex:
            raise ErrorInsertingNewUpdateVersion(ex)

    def get_last_loader_update_id(self):
        # type: () -> int

        def inner_func(conn):
            # type: (Connection) -> int

            query = """
                SELECT UpdateId
                FROM UpdatesController
                WHERE TypeId = '{}'
                ORDER BY Id DESC
                LIMIT 1""".format(self.update_value)

            cursor = conn.select(query)
            row = cursor.get_row(0)

            if row:
                update_id = row.get_entry("UpdateId")
                return int(update_id)

            return 0

        return self.execute_with_connection(inner_func, service_name=self.configs.persistence_name)

    def update_pos_ctrl_type(self, pos_id, user_control_type):
        # type: (int, int) -> None

        try:
            def inner_func(conn):
                # type: (Connection) -> None

                query = """
                    UPDATE PosState
                    SET UserControlType = '{0}'
                    WHERE PosId = '{1}'""".format(user_control_type, pos_id)
                conn.query(query)

            self.execute_with_connection(inner_func, service_name='Persistence')
        except Exception as ex:
            raise ErrorUpdatingUserControlType(ex)

    def get_pending_loader_update_id(self):
        # type: () -> int

        def inner_func(conn):
            # type: (Connection) -> int

            query = """
                SELECT UpdateId
                FROM UpdatesController
                WHERE TypeId = '{0}' AND StatusId = '{1}' AND AppliedDate IS NULL
                ORDER BY Id DESC""".format(self.update_value, UpdateStatus.pending)

            cursor = conn.select(query)
            row = cursor.get_row(0)

            if row:
                update_id = row.get_entry("UpdateId")
                return int(update_id)

            return 0

        return self.execute_with_connection(inner_func, service_name=self.configs.persistence_name)

    def mark_loader_update_as_applied(
            self,
    ):
        # type: (...) -> None

        try:
            def inner_func(conn):
                # type: (Connection) -> None

                date_now = str(datetime.now().isoformat())
                query = """
                        UPDATE UpdatesController
                        SET StatusId = '{0}', AppliedDate = '{1}', NotifiedDate = '{1}'
                        WHERE UpdateName = 'Loaders' AND AppliedDate IS NULL
                        """.format(UpdateStatus.applied, date_now)

                conn.query(query)

            self.execute_with_connection(inner_func, service_name=self.configs.persistence_name)
        except Exception as ex:
            raise ErrorUpdatingUserControlType(ex)

    def mark_loader_update_as_error(
            self,
    ):
        # type: (...) -> None

        try:
            def inner_func(conn):
                # type: (Connection) -> None

                query = """
                        UPDATE UpdatesController
                        SET StatusId = '{0}'
                        WHERE UpdateName = 'Loaders' AND AppliedDate IS NULL
                        """.format(UpdateStatus.error)

                conn.query(query)

            self.execute_with_connection(inner_func, service_name=self.configs.persistence_name)
        except Exception as ex:
            raise ErrorUpdatingUserControlType(ex)

    def has_already_applied_update(self):
        # type: () -> bool

        return self.common_pos_repository.has_already_applied_update(self.configs.persistence_name, self.update_value)

    def get_last_loader_update_date(self):
        # type: () -> Optional[datetime]

        return self.common_pos_repository.get_last_update_date(self.configs.persistence_name, self.update_value)

    def restart_hv(self):
        # type: () -> None
        res = restart_system(
            mb_context=self.message_bus.mbcontext,
            event_type="POS_UPDATER_LOADER",
        )
        if res.token != TK_SYS_ACK:
            raise ErrorRestartingHV("Could not perform HV global restart")

    def get_pos_users(self, pos_id):
        # type: (int) -> str

        message = Message(TK_POS_LISTUSERS, data_type=DataType.param, data="{}".format(pos_id))
        res = self.message_bus.send_message("POS{}".format(pos_id), message)
        if res.token != TK_SYS_ACK:
            raise ErrorGettingUsersList("Could not list users in POS{}".format(pos_id))
        return res.data

    def get_model(
        self,
        pos_id    # type: Any[Union[str, int]]
    ):
        # type: (...) -> etree.XML

        message = Message(TK_POS_GETMODEL, data_type=DataType.param, data="{}".format(pos_id))
        res = self.message_bus.send_message("POS{}".format(pos_id), message)
        if res.token != TK_SYS_ACK:
            raise ErrorGettingUsersList("Could not list users in POS{}".format(pos_id))
        return etree.XML(res.data)



# -*- coding: utf-8 -*-
# Module name: _threadpooltemplate.py
# Module Description: Thread pool template
#
# Copyright (C) 2018 MWneo Corporation
# Copyright (C) 2018 Omega Tech Enterprises Ltd. 
# (All rights transferred from MWneo Corporation to Omega Tech Enterprises Ltd.)
#
# $Id: _threadpooltemplate.py $
# $Revision: $
# $Date: $
# $Author: amerolli $

from Queue import Queue
from threading import Thread


class Worker(Thread):
    def __init__(self, queue):
        super(Worker, self).__init__()
        self._q = queue
        self.daemon = True
        self.start()

    def run(self):
        while True:
            f, args, kwargs = self._q.get()
            try:
                f(*args, **kwargs)
            except Exception as e:
                print(e)
            self._q.task_done()


class ThreadPool(object):
    def __init__(self, num_t=5):
        self._q = Queue(num_t)
        for _ in range(num_t):
            Worker(self._q)

    def add_task(self, f, *args, **kwargs):
        self._q.put((f, args, kwargs))
    
    def wait_complete(self):
        self._q.join()



def find_key_value(view_group, key_name, default_value=""):
    key = [x for x in view_group.keys if x.name == key_name]

    return str(default_value) if not key else key[0].values[0]


def find_key_values(view_group, key_name):
    key = [x for x in view_group.keys if x.name == key_name]

    return [] if not key else key[0].values



from xml.etree import cElementTree as eTree

from modelcommon.fiscalwrapper import (
    FiscalizationErrorsCode,
    FiscalizationException,
)
from mw_helper import format_currency

from typing import Dict, Tuple, Optional, List


class FiscalParameterNotFound(Exception):
    def __init__(self, part_code, parameter_name):
        # type: (str, unicode) -> None

        super(FiscalParameterNotFound, self).__init__("FiscalParameterNotFound")
        self.part_code = str(part_code)
        self.parameter_name = parameter_name

    def __str__(self):
        return "FiscalParameterNotFound: partCode: {0} - parameterName: {1}".format(self.part_code, self.parameter_name)


class FiscalParameterController(object):
    def __init__(
        self,
        fiscal_parameters,          # type: Dict[Tuple[str, unicode],unicode]
        tip_config_range,           # type: Dict
        loader_fiscal_parameters,   # type: List[Dict]
        fiscal_max_value,           # type: float
    ):
        # type: (...) -> None

        self.fiscal_parameters = fiscal_parameters
        self.tip_config_range = tip_config_range
        self.loader_fiscal_parameters = loader_fiscal_parameters
        self.fiscal_max_value = fiscal_max_value

    def get_optional_parameter(
            self,
            part_code,      # type: str
            parameter_name  # type: unicode
    ):
        # type: (...) -> Optional[unicode]

        key = (str(part_code), parameter_name)
        if key not in self.fiscal_parameters:
            return None
        else:
            return self.fiscal_parameters[key] or None

    def get_parameter(
            self,
            part_code,      # type: str
            parameter_name  # type: unicode
    ):
        # type: (...) -> unicode

        key = (str(part_code), parameter_name)
        if key not in self.fiscal_parameters:
            raise FiscalParameterNotFound(part_code, parameter_name)
        else:
            return self.fiscal_parameters[key] or None

    def has_parameter(
            self,
            part_code,      # type: str
            parameter_name  # type: unicode
    ):
        # type: (...) -> bool

        key = (str(part_code), parameter_name)
        return key in self.fiscal_parameters

    def get_all_products_csts(self):
        # type: () -> unicode

        return str({(x, y): self.fiscal_parameters.get((x, y)) for x, y in self.fiscal_parameters if "CST" in y})

    def get_all_products_base_reductions(self):
        # type: () -> unicode

        return str({x: self.fiscal_parameters.get((x, "BASE_REDUCTION")) for x, y in self.fiscal_parameters})

    def get_tip_params(self, part_code=None, taxes_name=None):
        # type: (str, List) -> []

        products = []
        try:
            if len(self.tip_config_range) == 0:
                return products
            for item in self.tip_config_range:
                if item["exempt"] or (part_code is not None and part_code != item["TipCode"]):
                    continue
                products.extend(self._get_taxes(item, taxes_name))
        except Exception as _:
            pass
        return products

    def _get_taxes(self, item, taxes_name=None):
        # type:(Dict, []) -> []

        taxes = ['PIS', 'COFINS', 'ICMS'] if taxes_name is None else taxes_name
        products = []
        for product in self.loader_fiscal_parameters:
            product_code = str(product['PRODUCT_CODE'])
            tip_code = str(item["TipCode"])
            if product_code != tip_code:
                continue
            for param in product:
                if param not in taxes:
                    continue
                rate = float(product[param])
                code = "{}_{}".format(str(param), rate)
                name = str(param)
                product_code = {tip_code: tip_code}
                index = str(param[:1])
                result = {"fiscalIndex": index, "code": code, "rate": rate, "name": name, "product": product_code}

                products.append(result)

        return products

    def validate_fiscal_max_value(
        self,   # type: FiscalParameterController
        model,  # type: eTree.XML
        order,  # type: eTree.XML
    ):
        # type: (...) -> Optional[str]

        try:
            self.can_fiscalize_order(order=order)
        except FiscalizationException as ex:
            formatted_value = format_currency(
                model=model,
                value=float(ex.message),
            )
            return "$MAXIMUM_AMOUNT_TO_FISCALIZE|{0}|{0}".format(formatted_value)

    def can_fiscalize_order(
        self,   # type: FiscalParameterController
        order,  # type: eTree.XML
    ):
        # type: (...) -> None

        order_value = float(order.get("totalGross") or "0.0") + float(order.get("tip") or "0.0")

        if order_value > self.fiscal_max_value:
            raise FiscalizationException(
                error_code=FiscalizationErrorsCode.FiscalMaxValueExceeded.value,
                message=str(self.fiscal_max_value)
            )



from ._DefaultManagerConfig import DefaultManagerConfig
from ._CatalogConfigurations import CatalogConfigurations
from ._MediaConfigurations import MediaConfigurations
from ._UserConfigurations import UserConfigurations
from ._Configurations import Configurations
from ._LoaderConfigurations import LoaderConfigurations
from ._BinaryConfigurations import BinaryConfigurations
from ._SystemPaths import SystemPaths



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import json
import logging
import re
import time

from botocore.compat import ensure_unicode, ensure_bytes, urlparse
from botocore.retryhandler import EXCEPTION_MAP as RETRYABLE_EXCEPTIONS


logger = logging.getLogger(__name__)


class Monitor(object):
    _EVENTS_TO_REGISTER = [
        'before-parameter-build',
        'request-created',
        'response-received',
        'after-call',
        'after-call-error',
    ]

    def __init__(self, adapter, publisher):
        """Abstraction for monitoring clients API calls

        :param adapter: An adapter that takes event emitter events
            and produces monitor events

        :param publisher: A publisher for generated monitor events
        """
        self._adapter = adapter
        self._publisher = publisher

    def register(self, event_emitter):
        """Register an event emitter to the monitor"""
        for event_to_register in self._EVENTS_TO_REGISTER:
            event_emitter.register_last(event_to_register, self.capture)

    def capture(self, event_name, **payload):
        """Captures an incoming event from the event emitter

        It will feed an event emitter event to the monitor's adaptor to create
        a monitor event and then publish that event to the monitor's publisher.
        """
        try:
            monitor_event = self._adapter.feed(event_name, payload)
            if monitor_event:
                self._publisher.publish(monitor_event)
        except Exception as e:
            logger.debug(
                'Exception %s raised by client monitor in handling event %s',
                e, event_name, exc_info=True)


class MonitorEventAdapter(object):
    def __init__(self, time=time.time):
        """Adapts event emitter events to produce monitor events

        :type time: callable
        :param time: A callable that produces the current time
        """
        self._time = time

    def feed(self, emitter_event_name, emitter_payload):
        """Feed an event emitter event to generate a monitor event

        :type emitter_event_name: str
        :param emitter_event_name: The name of the event emitted

        :type emitter_payload: dict
        :param emitter_payload: The payload to associated to the event
            emitted

        :rtype: BaseMonitorEvent
        :returns: A monitor event based on the event emitter events
            fired
        """
        return self._get_handler(emitter_event_name)(**emitter_payload)

    def _get_handler(self, event_name):
        return getattr(
            self, '_handle_' + event_name.split('.')[0].replace('-', '_')
        )

    def _handle_before_parameter_build(self, model, context, **kwargs):
        context['current_api_call_event'] = APICallEvent(
            service=model.service_model.service_id,
            operation=model.wire_name,
            timestamp=self._get_current_time(),
        )

    def _handle_request_created(self, request, **kwargs):
        context = request.context
        new_attempt_event = context[
            'current_api_call_event'].new_api_call_attempt(
                timestamp=self._get_current_time())
        new_attempt_event.request_headers = request.headers
        new_attempt_event.url = request.url
        context['current_api_call_attempt_event'] = new_attempt_event

    def _handle_response_received(self, parsed_response, context, exception,
                                  **kwargs):
        attempt_event = context.pop('current_api_call_attempt_event')
        attempt_event.latency = self._get_latency(attempt_event)
        if parsed_response is not None:
            attempt_event.http_status_code = parsed_response[
                'ResponseMetadata']['HTTPStatusCode']
            attempt_event.response_headers = parsed_response[
                'ResponseMetadata']['HTTPHeaders']
            attempt_event.parsed_error = parsed_response.get('Error')
        else:
            attempt_event.wire_exception = exception
        return attempt_event

    def _handle_after_call(self, context, parsed, **kwargs):
        context['current_api_call_event'].retries_exceeded = parsed[
            'ResponseMetadata'].get('MaxAttemptsReached', False)
        return self._complete_api_call(context)

    def _handle_after_call_error(self, context, exception, **kwargs):
        # If the after-call-error was emitted and the error being raised
        # was a retryable connection error, then the retries must have exceeded
        # for that exception as this event gets emitted **after** retries
        # happen.
        context['current_api_call_event'].retries_exceeded = \
            self._is_retryable_exception(exception)
        return self._complete_api_call(context)

    def _is_retryable_exception(self, exception):
        return isinstance(
            exception, tuple(RETRYABLE_EXCEPTIONS['GENERAL_CONNECTION_ERROR']))

    def _complete_api_call(self, context):
        call_event = context.pop('current_api_call_event')
        call_event.latency = self._get_latency(call_event)
        return call_event

    def _get_latency(self, event):
        return self._get_current_time() - event.timestamp

    def _get_current_time(self):
        return int(self._time() * 1000)


class BaseMonitorEvent(object):
    def __init__(self, service, operation, timestamp):
        """Base monitor event

        :type service: str
        :param service: A string identifying the service associated to
            the event

        :type operation: str
        :param operation: A string identifying the operation of service
            associated to the event

        :type timestamp: int
        :param timestamp: Epoch time in milliseconds from when the event began
        """
        self.service = service
        self.operation = operation
        self.timestamp = timestamp

    def __repr__(self):
        return '%s(%r)' % (self.__class__.__name__, self.__dict__)

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.__dict__ == other.__dict__
        return False


class APICallEvent(BaseMonitorEvent):
    def __init__(self, service, operation, timestamp, latency=None,
                 attempts=None, retries_exceeded=False):
        """Monitor event for a single API call

        This event corresponds to a single client method call, which includes
        every HTTP requests attempt made in order to complete the client call

        :type service: str
        :param service: A string identifying the service associated to
            the event

        :type operation: str
        :param operation: A string identifying the operation of service
            associated to the event

        :type timestamp: int
        :param timestamp: Epoch time in milliseconds from when the event began

        :type latency: int
        :param latency: The time in milliseconds to complete the client call

        :type attempts: list
        :param attempts: The list of APICallAttempts associated to the
            APICall

        :type retries_exceeded: bool
        :param retries_exceeded: True if API call exceeded retries. False
            otherwise
        """
        super(APICallEvent, self).__init__(
            service=service, operation=operation, timestamp=timestamp)
        self.latency = latency
        self.attempts = attempts
        if attempts is None:
            self.attempts = []
        self.retries_exceeded = retries_exceeded

    def new_api_call_attempt(self, timestamp):
        """Instantiates APICallAttemptEvent associated to the APICallEvent

        :type timestamp: int
        :param timestamp: Epoch time in milliseconds to associate to the
            APICallAttemptEvent
        """
        attempt_event = APICallAttemptEvent(
            service=self.service,
            operation=self.operation,
            timestamp=timestamp
        )
        self.attempts.append(attempt_event)
        return attempt_event


class APICallAttemptEvent(BaseMonitorEvent):
    def __init__(self, service, operation, timestamp,
                 latency=None, url=None, http_status_code=None,
                 request_headers=None, response_headers=None,
                 parsed_error=None, wire_exception=None):
        """Monitor event for a single API call attempt

        This event corresponds to a single HTTP request attempt in completing
        the entire client method call.

        :type service: str
        :param service: A string identifying the service associated to
            the event

        :type operation: str
        :param operation: A string identifying the operation of service
            associated to the event

        :type timestamp: int
        :param timestamp: Epoch time in milliseconds from when the HTTP request
            started

        :type latency: int
        :param latency: The time in milliseconds to complete the HTTP request
            whether it succeeded or failed

        :type url: str
        :param url: The URL the attempt was sent to

        :type http_status_code: int
        :param http_status_code: The HTTP status code of the HTTP response
            if there was a response

        :type request_headers: dict
        :param request_headers: The HTTP headers sent in making the HTTP
            request

        :type response_headers: dict
        :param response_headers: The HTTP headers returned in the HTTP response
            if there was a response

        :type parsed_error: dict
        :param parsed_error: The error parsed if the service returned an
            error back

        :type wire_exception: Exception
        :param wire_exception: The exception raised in sending the HTTP
            request (i.e. ConnectionError)
        """
        super(APICallAttemptEvent, self).__init__(
            service=service, operation=operation, timestamp=timestamp
        )
        self.latency = latency
        self.url = url
        self.http_status_code = http_status_code
        self.request_headers = request_headers
        self.response_headers = response_headers
        self.parsed_error = parsed_error
        self.wire_exception = wire_exception


class CSMSerializer(object):
    _MAX_CLIENT_ID_LENGTH = 255
    _MAX_EXCEPTION_CLASS_LENGTH = 128
    _MAX_ERROR_CODE_LENGTH = 128
    _MAX_USER_AGENT_LENGTH = 256
    _MAX_MESSAGE_LENGTH = 512
    _RESPONSE_HEADERS_TO_EVENT_ENTRIES = {
        'x-amzn-requestid': 'XAmznRequestId',
        'x-amz-request-id': 'XAmzRequestId',
        'x-amz-id-2': 'XAmzId2',
    }
    _AUTH_REGEXS = {
        'v4': re.compile(
            r'AWS4-HMAC-SHA256 '
            r'Credential=(?P<access_key>\w+)/\d+/'
            r'(?P<signing_region>[a-z0-9-]+)/'
        ),
        's3': re.compile(
            r'AWS (?P<access_key>\w+):'
        )
    }
    _SERIALIZEABLE_EVENT_PROPERTIES = [
        'service',
        'operation',
        'timestamp',
        'attempts',
        'latency',
        'retries_exceeded',
        'url',
        'request_headers',
        'http_status_code',
        'response_headers',
        'parsed_error',
        'wire_exception',
    ]

    def __init__(self, csm_client_id):
        """Serializes monitor events to CSM (Client Side Monitoring) format

        :type csm_client_id: str
        :param csm_client_id: The application identifier to associate
            to the serialized events
        """
        self._validate_client_id(csm_client_id)
        self.csm_client_id = csm_client_id

    def _validate_client_id(self, csm_client_id):
        if len(csm_client_id) > self._MAX_CLIENT_ID_LENGTH:
            raise ValueError(
                'The value provided for csm_client_id: %s exceeds the '
                'maximum length of %s characters' % (
                    csm_client_id, self._MAX_CLIENT_ID_LENGTH)
            )

    def serialize(self, event):
        """Serializes a monitor event to the CSM format

        :type event: BaseMonitorEvent
        :param event: The event to serialize to bytes

        :rtype: bytes
        :returns: The CSM serialized form of the event
        """
        event_dict = self._get_base_event_dict(event)
        event_type = self._get_event_type(event)
        event_dict['Type'] = event_type
        for attr in self._SERIALIZEABLE_EVENT_PROPERTIES:
            value = getattr(event, attr, None)
            if value is not None:
                getattr(self, '_serialize_' + attr)(
                    value, event_dict, event_type=event_type)
        return ensure_bytes(
            json.dumps(event_dict, separators=(',', ':')))

    def _get_base_event_dict(self, event):
        return {
            'Version': 1,
            'ClientId': self.csm_client_id,
        }

    def _serialize_service(self, service, event_dict, **kwargs):
        event_dict['Service'] = service

    def _serialize_operation(self, operation, event_dict, **kwargs):
        event_dict['Api'] = operation

    def _serialize_timestamp(self, timestamp, event_dict, **kwargs):
        event_dict['Timestamp'] = timestamp

    def _serialize_attempts(self, attempts, event_dict, **kwargs):
        event_dict['AttemptCount'] = len(attempts)
        if attempts:
            self._add_fields_from_last_attempt(event_dict, attempts[-1])

    def _add_fields_from_last_attempt(self, event_dict, last_attempt):
        if last_attempt.request_headers:
            # It does not matter which attempt to use to grab the region
            # for the ApiCall event, but SDKs typically do the last one.
            region = self._get_region(last_attempt.request_headers)
            if region is not None:
                event_dict['Region'] = region
            event_dict['UserAgent'] = self._get_user_agent(
                last_attempt.request_headers)
        if last_attempt.http_status_code is not None:
            event_dict['FinalHttpStatusCode'] = last_attempt.http_status_code
        if last_attempt.parsed_error is not None:
            self._serialize_parsed_error(
                last_attempt.parsed_error, event_dict, 'ApiCall')
        if last_attempt.wire_exception is not None:
            self._serialize_wire_exception(
                last_attempt.wire_exception, event_dict, 'ApiCall')

    def _serialize_latency(self, latency, event_dict, event_type):
        if event_type == 'ApiCall':
            event_dict['Latency'] = latency
        elif event_type == 'ApiCallAttempt':
            event_dict['AttemptLatency'] = latency

    def _serialize_retries_exceeded(self, retries_exceeded, event_dict,
                                    **kwargs):
        event_dict['MaxRetriesExceeded'] = (1 if retries_exceeded else 0)

    def _serialize_url(self, url, event_dict, **kwargs):
        event_dict['Fqdn'] = urlparse(url).netloc

    def _serialize_request_headers(self, request_headers, event_dict,
                                   **kwargs):
        event_dict['UserAgent'] = self._get_user_agent(request_headers)
        if self._is_signed(request_headers):
            event_dict['AccessKey'] = self._get_access_key(request_headers)
        region = self._get_region(request_headers)
        if region is not None:
            event_dict['Region'] = region
        if 'X-Amz-Security-Token' in request_headers:
            event_dict['SessionToken'] = request_headers[
                'X-Amz-Security-Token']

    def _serialize_http_status_code(self, http_status_code, event_dict,
                                    **kwargs):
        event_dict['HttpStatusCode'] = http_status_code

    def _serialize_response_headers(self, response_headers, event_dict,
                                    **kwargs):
        for header, entry in self._RESPONSE_HEADERS_TO_EVENT_ENTRIES.items():
            if header in response_headers:
                event_dict[entry] = response_headers[header]

    def _serialize_parsed_error(self, parsed_error, event_dict, event_type,
                                **kwargs):
        field_prefix = 'Final' if event_type == 'ApiCall' else ''
        event_dict[field_prefix + 'AwsException'] = self._truncate(
            parsed_error['Code'], self._MAX_ERROR_CODE_LENGTH)
        event_dict[field_prefix + 'AwsExceptionMessage'] = self._truncate(
            parsed_error['Message'], self._MAX_MESSAGE_LENGTH)

    def _serialize_wire_exception(self, wire_exception, event_dict, event_type,
                                  **kwargs):
        field_prefix = 'Final' if event_type == 'ApiCall' else ''
        event_dict[field_prefix + 'SdkException'] = self._truncate(
            wire_exception.__class__.__name__,
            self._MAX_EXCEPTION_CLASS_LENGTH)
        event_dict[field_prefix + 'SdkExceptionMessage'] = self._truncate(
            str(wire_exception), self._MAX_MESSAGE_LENGTH)

    def _get_event_type(self, event):
        if isinstance(event, APICallEvent):
            return 'ApiCall'
        elif isinstance(event, APICallAttemptEvent):
            return 'ApiCallAttempt'

    def _get_access_key(self, request_headers):
        auth_val = self._get_auth_value(request_headers)
        _, auth_match = self._get_auth_match(auth_val)
        return auth_match.group('access_key')

    def _get_region(self, request_headers):
        if not self._is_signed(request_headers):
            return None
        auth_val = self._get_auth_value(request_headers)
        signature_version, auth_match = self._get_auth_match(auth_val)
        if signature_version != 'v4':
            return None
        return auth_match.group('signing_region')

    def _get_user_agent(self, request_headers):
        return self._truncate(
            ensure_unicode(request_headers.get('User-Agent', '')),
            self._MAX_USER_AGENT_LENGTH
        )

    def _is_signed(self, request_headers):
        return 'Authorization' in request_headers

    def _get_auth_value(self, request_headers):
        return ensure_unicode(request_headers['Authorization'])

    def _get_auth_match(self, auth_val):
        for signature_version, regex in self._AUTH_REGEXS.items():
            match = regex.match(auth_val)
            if match:
                return signature_version, match
        return None, None

    def _truncate(self, text, max_length):
        if len(text) > max_length:
            logger.debug(
                'Truncating following value to maximum length of '
                '%s: %s', text, max_length)
            return text[:max_length]
        return text


class SocketPublisher(object):
    _MAX_MONITOR_EVENT_LENGTH = 8 * 1024

    def __init__(self, socket, host, port, serializer):
        """Publishes monitor events to a socket

        :type socket: socket.socket
        :param socket: The socket object to use to publish events

        :type host: string
        :param host: The host to send events to

        :type port: integer
        :param port: The port on the host to send events to

        :param serializer: The serializer to use to serialize the event
            to a form that can be published to the socket. This must
            have a `serialize()` method that accepts a monitor event
            and return bytes
        """
        self._socket = socket
        self._address = (host, port)
        self._serializer = serializer

    def publish(self, event):
        """Publishes a specified monitor event

        :type event: BaseMonitorEvent
        :param event: The monitor event to be sent
            over the publisher's socket to the desired address.
        """
        serialized_event = self._serializer.serialize(event)
        if len(serialized_event) > self._MAX_MONITOR_EVENT_LENGTH:
            logger.debug(
                'Serialized event of size %s exceeds the maximum length '
                'allowed: %s. Not sending event to socket.',
                len(serialized_event), self._MAX_MONITOR_EVENT_LENGTH
            )
            return
        self._socket.sendto(serialized_event, self._address)



from application.domain.ui import InfoMessage
from sysactions import show_info_message


class MsgBusInfoMessage(InfoMessage):
    def __init__(self, pos_id):
        self.pos_id = pos_id

    def show(self, message, msg_type=None):
        show_info_message(self.pos_id, message, msgtype=msg_type.name)



def migrate(conn, tables):
    updateTablesPriority = [
        "FiscalPrinters", "FiscalOrders", "FiscalOrderItems",
        "FiscalOrderTender", "NonFiscalDocuments", "NonFiscalDocumentTenders",
        "ZTapes", "ZTapeTotalizers", "NotasManuais", "ElectronicTransactions",
        "Estoque", "EstoqueHistorico", "Insumos"
    ]
    conn.query("DELETE FROM FiscalDEnabled;")
    for table in updateTablesPriority:
        strCommonFields = ", ".join(tables[table]["fields_Common"])
        stmt = "INSERT OR REPLACE INTO %s (%s) SELECT %s FROM old.%s;" % (table, strCommonFields, strCommonFields, table)
        conn.query(stmt)
    conn.query("INSERT INTO FiscalDEnabled(Enabled) VALUES(1);")



# -*- coding: utf-8 -*-
r"""
    werkzeug.posixemulation
    ~~~~~~~~~~~~~~~~~~~~~~~

    Provides a POSIX emulation for some features that are relevant to
    web applications.  The main purpose is to simplify support for
    systems such as Windows NT that are not 100% POSIX compatible.

    Currently this only implements a :func:`rename` function that
    follows POSIX semantics.  Eg: if the target file already exists it
    will be replaced without asking.

    This module was introduced in 0.6.1 and is not a public interface.
    It might become one in later versions of Werkzeug.

    :copyright: 2007 Pallets
    :license: BSD-3-Clause
"""
import errno
import os
import random
import sys
import time

from ._compat import to_unicode
from .filesystem import get_filesystem_encoding

can_rename_open_file = False

if os.name == "nt":
    try:
        import ctypes

        _MOVEFILE_REPLACE_EXISTING = 0x1
        _MOVEFILE_WRITE_THROUGH = 0x8
        _MoveFileEx = ctypes.windll.kernel32.MoveFileExW

        def _rename(src, dst):
            src = to_unicode(src, get_filesystem_encoding())
            dst = to_unicode(dst, get_filesystem_encoding())
            if _rename_atomic(src, dst):
                return True
            retry = 0
            rv = False
            while not rv and retry < 100:
                rv = _MoveFileEx(
                    src, dst, _MOVEFILE_REPLACE_EXISTING | _MOVEFILE_WRITE_THROUGH
                )
                if not rv:
                    time.sleep(0.001)
                    retry += 1
            return rv

        # new in Vista and Windows Server 2008
        _CreateTransaction = ctypes.windll.ktmw32.CreateTransaction
        _CommitTransaction = ctypes.windll.ktmw32.CommitTransaction
        _MoveFileTransacted = ctypes.windll.kernel32.MoveFileTransactedW
        _CloseHandle = ctypes.windll.kernel32.CloseHandle
        can_rename_open_file = True

        def _rename_atomic(src, dst):
            ta = _CreateTransaction(None, 0, 0, 0, 0, 1000, "Werkzeug rename")
            if ta == -1:
                return False
            try:
                retry = 0
                rv = False
                while not rv and retry < 100:
                    rv = _MoveFileTransacted(
                        src,
                        dst,
                        None,
                        None,
                        _MOVEFILE_REPLACE_EXISTING | _MOVEFILE_WRITE_THROUGH,
                        ta,
                    )
                    if rv:
                        rv = _CommitTransaction(ta)
                        break
                    else:
                        time.sleep(0.001)
                        retry += 1
                return rv
            finally:
                _CloseHandle(ta)

    except Exception:

        def _rename(src, dst):
            return False

        def _rename_atomic(src, dst):
            return False

    def rename(src, dst):
        # Try atomic or pseudo-atomic rename
        if _rename(src, dst):
            return
        # Fall back to "move away and replace"
        try:
            os.rename(src, dst)
        except OSError as e:
            if e.errno != errno.EEXIST:
                raise
            old = "%s-%08x" % (dst, random.randint(0, sys.maxsize))
            os.rename(dst, old)
            os.rename(src, dst)
            try:
                os.unlink(old)
            except Exception:
                pass


else:
    rename = os.rename
    can_rename_open_file = True



class OrderData(object):
    def __init__(self, order_id, pos_id, order_picture, order_xml):
        # type: (int, int, unicode, unicode) -> None
        self.order_id = order_id
        self.pos_id = pos_id
        self.order_picture = order_picture
        self.order_xml = order_xml



# -*- coding: utf-8 -*-
from helper import unicode_2_ascii

from ibpttaxprocessor import IbptTaxProcessor
from old_helper import round_half_away_from_zero
from typing import Dict, Any
from xml.sax.saxutils import escape
from pos_model import Order, TaxItem, SaleItem

from common import FiscalParameterController
from pos_util import SaleLineUtil

from _SatXmlPartBuilder import SatXmlPartBuilder
from _ContextKeys import ContextKeys
from _TaxBuilder import TaxBuilderLucroReal, TaxBuilderSimplesNacional, TaxBuilderSimplesNacionalSublimiteReceita


class DetBuilder(SatXmlPartBuilder):
    def __init__(self, crt, fiscal_parameter_controller, sale_line_util, ibtp_tax_processor):
        # type: (int, FiscalParameterController, SaleLineUtil, IbptTaxProcessor) -> None
        self.crt = crt
        self.fiscal_parameter_controller = fiscal_parameter_controller
        self.sale_line_util = sale_line_util
        self.ibtp_tax_processor = ibtp_tax_processor
        if crt == 1:
            self.tax_builder = TaxBuilderSimplesNacional()
        elif crt == 2:
            self.tax_builder = TaxBuilderSimplesNacionalSublimiteReceita()
        elif crt == 3:
            self.tax_builder = TaxBuilderLucroReal()
        else:
            raise Exception("CRT Não informado")

    def build_xml(self, order, context):
        # type: (Order, Dict[unicode, Any]) -> unicode

        item_no = context[ContextKeys.item_no]  # type: int
        sale_item = context[ContextKeys.current_sale_item]  # type: SaleItem

        product_code = sale_item.part_code
        cest = self.fiscal_parameter_controller.get_optional_parameter(product_code, "CEST")

        ncm = self.fiscal_parameter_controller.get_parameter(product_code, "NCM")
        if ncm is None:
            raise Exception("NCM Não encontrado\\%s-%s" % (product_code, sale_item.product_name))

        cfop = self.fiscal_parameter_controller.get_parameter(product_code, "CFOP")
        if cfop is None:
            raise Exception("CFOP Não encontrado\\%s-%s" % (product_code, sale_item.product_name))

        csosn = None
        if self.crt == 1:
            csosn = self.fiscal_parameter_controller.get_parameter(product_code, "CSOSN")
            if csosn is None:
                raise Exception("CSOSN Não encontrado\\%s-%s" % (product_code, sale_item.product_name))
            csosn = int(csosn)

        cst_icms = self.fiscal_parameter_controller.get_parameter(product_code, "CST_ICMS")
        if cst_icms is None:
            raise Exception("CST ICMS Não encontrado\\%s-%s" % (product_code, sale_item.product_name))
        cst_icms = int(cst_icms)

        cst_pis = self.fiscal_parameter_controller.get_parameter(product_code, "CST_PIS")
        if cst_pis is None:
            raise Exception("CST PIS Não encontrado\\%s-%s" % (product_code, sale_item.product_name))
        cst_pis = int(cst_pis)

        cst_cofins = self.fiscal_parameter_controller.get_parameter(product_code, "CST_COFINS")
        if cst_cofins is None:
            raise Exception("CST COFINS Não encontrado\\%s-%s" % (product_code, sale_item.product_name))
        cst_cofins = int(cst_cofins)

        base_reduction = float(self.fiscal_parameter_controller.get_optional_parameter(product_code, "BASE_REDUCTION") or 0)
        product_name = unicode_2_ascii(escape(sale_item.product_name))
        cest_xml = "" if cest is None else "<obsFiscoDet xCampoDet='Cod.CEST'><xTextoDet>{x_text_det:>07}</xTextoDet></obsFiscoDet>".format(x_text_det=cest)

        prod_xml_template = u"<prod>" \
                            "<cProd>{c_prod}</cProd>" \
                            "<xProd>{x_prod}</xProd>" \
                            "<NCM>{ncm:0>8}</NCM>" \
                            "<CFOP>{cfop}</CFOP>" \
                            "<uCom>{measure_unit}</uCom>" \
                            "<qCom>{q_com:0.4f}</qCom>" \
                            "<vUnCom>{v_un_com:0.3f}</vUnCom>" \
                            "<indRegra>A</indRegra>"\
                            "{discount_xml}"\
                            "{cest_xml}" \
                            "</prod>"

        item_discount = sale_item.item_discount or 0
        discount_xml = ""
        discount_value = 0 if item_discount < 0 else item_discount
        increase_value = item_discount if item_discount < 0 else 0
        discount_xml += "<{0}>{1:.02f}</{0}>".format("vDesc",  abs(discount_value))
        discount_xml += "<{0}>{1:.02f}</{0}>".format("vOutro", abs(increase_value))

        item_value = round_half_away_from_zero(sale_item.item_price / sale_item.quantity, 2)

        prod_xml = prod_xml_template.format(
            c_prod=product_code,
            x_prod=product_name.strip(),
            ncm=int(ncm),
            cfop=cfop,
            q_com=sale_item.quantity,
            v_un_com=item_value,
            discount_xml=discount_xml,
            cest_xml=cest_xml,
            measure_unit=sale_item.measure_unit.upper()
        )

        impost_xml_template = "<imposto>{tax_xml}</imposto>"

        v_item_12741 = self.tax_builder.calculate_approximated_tax_rate(self.ibtp_tax_processor, sale_item)
        tax_xml = v_item_12741

        icms_tax_item = sale_item.get_tax_item("ICMS")  # type: TaxItem
        xml, vbc, icms_tax = self.tax_builder.calculate_icms(cst_icms, csosn, icms_tax_item, base_reduction, product_code)
        tax_xml += xml
        parcial_vbc = vbc

        pis_tax_item = sale_item.get_tax_item("PIS")  # type: TaxItem
        xml, vbc, pis_tax = self.tax_builder.calculate_pis(cst_pis, pis_tax_item)
        tax_xml += xml

        cofins_tax_item = sale_item.get_tax_item("COFINS")  # type: TaxItem
        xml, vbc, cofins_tax = self.tax_builder.calculate_cofins(cst_cofins, cofins_tax_item)
        tax_xml += xml

        imposto_xml = impost_xml_template.format(tax_xml=tax_xml)

        xml = u"<det nItem=\"{0}\">{prod_xml}{imposto_xml}</det>".format(
            item_no,
            prod_xml=prod_xml,
            imposto_xml=imposto_xml
        )

        context[ContextKeys.total_cofins] += cofins_tax
        context[ContextKeys.total_icms] += icms_tax
        context[ContextKeys.total_pis] += pis_tax
        context[ContextKeys.total_prod] += sale_item.item_price
        context[ContextKeys.total_vbc] += parcial_vbc

        return xml



import errno
import json
import os
import shutil
import time
from datetime import datetime
from threading import Lock
from typing import Dict
import xml.etree.ElementTree as eTree

from application.model import UpdateType, OperatorStatus
from application.model.configuration import Configurations
from application.model.customexception import NotExpectedException, DisabledFunctionality, ErrorCreatingBackup, \
    ErrorDeletingContent, ErrorRecoveringFolder, ErrorExtractingLoaders, ErrorDownloadingLoaders, RollbackError, \
    NotInsideUpdateWindow, ErrorValidatingOpenedUsers, ErrorUpdatingUserRoles
from application.repository.api import LoaderUpdaterAPIRepository
from application.repository.pos import LoaderUpdaterPOSRepository, VersionsPOSRepository
from application.service import UpdaterService, StaticConfigurationService
from application.util import LoggerUtil, ValidationUtil, settings
from systools import sys_parsefilepath

GENESIS_PATH = "../../genesis"

logger = LoggerUtil.get_logger_name(__name__)


class LoaderUpdaterService(object):
    def __init__(self, configs, working_on_update, api_repository, pos_repository, version_repository, updater_service):
        # type: (Configurations, Lock, LoaderUpdaterAPIRepository, LoaderUpdaterPOSRepository, VersionsPOSRepository, UpdaterService) -> None # noqa

        self.update_type = UpdateType.loader
        self.update_name = self.update_type.name
        self.enabled = configs.updaters[self.update_name].enabled
        self.local_configs = configs.updaters[self.update_name]
        self.store_id = configs.store_id

        self.working_on_update = working_on_update
        self.api_repository = api_repository
        self.pos_repository = pos_repository
        self.version_repository = version_repository
        self.updater_service = updater_service

        data_path = sys_parsefilepath("{HVDATADIR}")
        work_dir = sys_parsefilepath(configs.backup_directory)
        self.data_relative_path = "data"
        self.server_relative_path = "data/server"
        self.databases_relative_path = "data/server/databases"
        self.data_path = os.path.join(data_path, GENESIS_PATH, self.data_relative_path)
        self.server_path = os.path.join(self.data_path, "server")
        self.databases_path = os.path.join(data_path, GENESIS_PATH, self.databases_relative_path)
        default_work_path = os.path.join(work_dir, self.update_type.name)
        self.downloads_path = os.path.join(default_work_path, "downloads")
        self.backups_path = os.path.join(default_work_path, "backups")
        self.work_path = os.path.join(default_work_path, "work_path")
        self.payments_json = os.path.join(self.server_path, "bundles/payments.json")
        self.genesis_operator_levels_json = os.path.join(self.server_path, "bundles/operatorLevels.json")
        self.htdocs_path = os.path.join(data_path, 'htdocs')
        self.htdocs_kui_path = os.path.join(self.htdocs_path, 'kui')
        self.sui_static_path = os.path.join(data_path, 'htdocs/sui/static')
        self.loaders_zip = "loaders.zip"

        if not os.path.exists(self.backups_path):
            os.makedirs(self.backups_path)
        if not os.path.exists(self.downloads_path):
            os.makedirs(self.downloads_path)

    def update_loaders(
        self,
        restart_hv=False,       # type: bool
        retry_on_error=False,   # type: bool
    ):
        # type: (...) -> None

        settings.set_update_control(self.update_name, None)
        if not self.enabled:
            settings.set_update_control(self.update_name, "$UPDATE_DISABLED")
            raise DisabledFunctionality()

        rollback_necessary = False
        backup_package_zip = ''

        try:
            logger.info("Waiting Updating Loaders")
            with self.working_on_update:
                logger.info("Updating loaders")
                settings.set_update_control(self.update_name, "$UPDATING_LOADERS")
                settings.set_update_control("updating", True)
                last_loader_update_id = self.pos_repository.get_last_loader_update_id()

                self._delete_directories()
                self.updater_service.create_dir(self.work_path)
                settings.set_update_control(self.update_name, "$GETTING_NEW_LOADERS")
                self._download_new_loaders()
                self.pos_repository.insert_new_loader_update(str(last_loader_update_id + 1), "Loaders")

                settings.set_update_control(self.update_name, "$LOADERS_BACKUP")
                backup_package_zip = self._create_backup()

                rollback_necessary = True

                settings.set_update_control(self.update_name, "$APPLYING_NEW_LOADERS")
                self._get_new_loaders()

                current_pos_ctrl_dir = os.path.join(self.work_path, self.server_relative_path, 'bundles', 'posctrl')
                changed_pos_ctrl_type = None
                if os.path.exists(current_pos_ctrl_dir):
                    changed_pos_ctrl_type = self._could_apply_new_loaders()

                self._recover_folders()

                settings.set_update_control(self.update_name, "$UPDATING_USER_FUNCTIONS")
                self.update_operator_levels()

                self.updater_service.update_databases_with_json_config(
                    update_name=self.update_name,
                    settings=settings,
                )

                self.pos_repository.mark_loader_update_as_applied()

                StaticConfigurationService.update_static_loaders()
                StaticConfigurationService.run_databases_script()
                settings.set_update_control(self.update_name, "$SUCCESS_UPDATE")

                if changed_pos_ctrl_type is not None:
                    self._update_pos_ctrl_type(changed_pos_ctrl_type)

                self._update_ui_customizations()

                rollback_necessary = False
                if restart_hv:
                    self._delete_directories()
                    logger.info("Starting Hypervisor to apply new Loaders")
                    self.pos_repository.restart_hv()
                    time.sleep(5)

                logger.info("Loaders update fully concluded with success!")
        except NotInsideUpdateWindow as _:
            settings.set_update_control(self.update_name, "$NOT_INSIDE_UPDATE_WINDOW")
            logger.info("Sleeping until enter the update window")
            raise
        except (NotExpectedException, ErrorDownloadingLoaders, Exception) as ex:
            if rollback_necessary:
                with self.working_on_update:
                    self._rollback(backup_package_zip)

            if isinstance(ex, ErrorDownloadingLoaders):
                settings.set_update_control(self.update_name, ex.message)
            else:
                settings.set_update_control(self.update_name, "$UNKNOWN_ERROR")

            if not retry_on_error:
                self.pos_repository.mark_loader_update_as_error()

            logger.exception("Not expected exception on loaders update")
            raise
        finally:
            self._delete_directories()
            settings.set_update_control("updating", False)

    def _update_pos_ctrl_type(self, pos_ctrl_type):
        # type: (Dict) -> None

        for pos_id, value in pos_ctrl_type.items():
            user_control_type = 1 if value == "QSR" else 2
            self.pos_repository.update_pos_ctrl_type(pos_id, user_control_type)

    def update_operator_levels(
            self
    ):
        # type: () -> None

        try:
            if os.path.exists(self.genesis_operator_levels_json) and os.path.exists(self.sui_static_path):
                current_operator_levels = os.path.join(self.sui_static_path, 'operatorLevels.json')

                with open(self.genesis_operator_levels_json, 'rb') as f:
                    new_operator_level = json.load(f)

                has_some_configuration = sum([len(x) for x in new_operator_level["levels"].values()]) > 0
                if has_some_configuration:
                    self._create_support_menu_buttons(
                        operator_level=new_operator_level
                    )
                    self.updater_service.save_json(
                        data=new_operator_level,
                        destination_path=current_operator_levels
                    )
                else:
                    self.updater_service.delete_file(
                        path=current_operator_levels
                    )
        except Exception as e:
            logger.info("Error processing operatorLevels.json: {}".format(e))
            raise ErrorUpdatingUserRoles()

    @staticmethod
    def _create_support_menu_buttons(
        operator_level,     # type: Dict
    ):
        # type: (...) -> None

        rows = int(operator_level["rows"])
        support_menu = {
            "function": "SUPPORT_MENU",
            "row": rows,
            "column": 0,
            "color": "#274E13"
        }
        update_users = {
            "function": "UPDATE_USERS",
            "row": rows,
            "column": 1,
            "color": "#274E13"
        }

        operator_level["levels"]["30"] = operator_level["levels"]["10"][:]
        operator_level["levels"]["30"].extend([support_menu, update_users])

    @staticmethod
    def _add_value_to_pos_dict(
        element_name,   # type: str
        xml_path,       # type: str
        pos_dict        # type: Dict
    ):
        # type: (...) -> None

        xml_loader = eTree.parse(xml_path)
        element_pos = xml_loader.find(".//*[@name='POS']/group")
        pos = element_pos.get("name") if element_pos is not None else None
        element = xml_loader.find(".//*[@name='{}']/string".format(element_name))
        element_value = element.text if element is not None else None
        if pos and element_value:
            pos_dict[pos] = element_value

    def _validate_update_pos_values(
        self,
        current_pos_values,     # type: Dict
        new_pos_values,         # type: Dict
        key_index,              # type: str
        validate_state=True     # type: bool
    ):
        # type: (...) -> Dict

        update_pos = {}
        for pos, value in new_pos_values.items():
            if pos not in current_pos_values.keys() or value == current_pos_values[pos]:
                continue

            pos_id = int(pos[3:])
            model = self.pos_repository.get_model(pos_id=pos_id)
            self._validate_users(model=model)
            if validate_state:
                self._validate_pos_state(model=model)
            logger.info("{}: has a different values for {}".format(pos, key_index))
            update_pos[pos_id] = value

        return update_pos

    @staticmethod
    def _validate_users(
        model  # type: eTree.XML
    ):
        # type: (...) -> None

        operators = model.find("Operators").findall("Operator")
        for operator in operators:
            status = operator.get("state")
            if status != "LOGGEDOUT":
                logger.exception("can't update loader with operator status: {}".format(status))
                raise ErrorValidatingOpenedUsers("NEED_TO_CLOSE_OPENED_USERS")

    def _could_apply_new_loaders(self):
        # type: () -> Dict

        current_user_control_type = {}
        new_user_control_type = {}
        current_pod_type = {}
        new_pod_type = {}
        new_pos_ctrl_dir = os.path.join(self.server_path, 'bundles', 'posctrl')
        current_pos_ctrl_dir = os.path.join(self.work_path, self.server_relative_path, 'bundles', 'posctrl')
        if not os.path.exists(new_pos_ctrl_dir) or not os.path.exists(current_pos_ctrl_dir):
            return {}

        current_loaders_paths = os.listdir(current_pos_ctrl_dir) or []
        current_loaders_paths = [x for x in current_loaders_paths if ".cfg" in x]
        for path in current_loaders_paths:
            self._add_value_to_pos_dict(
                element_name="userControlType",
                xml_path=os.path.join(current_pos_ctrl_dir, path),
                pos_dict=current_user_control_type
            )
            self._add_value_to_pos_dict(
                element_name="pod",
                xml_path=os.path.join(current_pos_ctrl_dir, path),
                pos_dict=current_pod_type
            )

        new_loaders_paths = os.listdir(new_pos_ctrl_dir) or []
        new_loaders_paths = [x for x in new_loaders_paths if ".cfg" in x]
        for path in new_loaders_paths:
            self._add_value_to_pos_dict(
                element_name="userControlType",
                xml_path=os.path.join(new_pos_ctrl_dir, path),
                pos_dict=new_user_control_type
            )
            self._add_value_to_pos_dict(
                element_name="pod",
                xml_path=os.path.join(new_pos_ctrl_dir, path),
                pos_dict=new_pod_type
            )

        self._validate_pod_type(
            current_pod_type=current_pod_type,
            new_pod_type=new_pod_type
        )
        return self._validate_user_control_type(
            current_user_control_type=current_user_control_type,
            new_user_control_type=new_user_control_type
        )

    def _validate_user_control_type(
        self,
        current_user_control_type,  # type: Dict
        new_user_control_type       # type: Dict
    ):
        # type: (...) -> Dict

        return self._validate_update_pos_values(
            current_pos_values=current_user_control_type,
            new_pos_values=new_user_control_type,
            key_index="user_control_type"
        )

    def _validate_pod_type(
        self,
        current_pod_type,   # type: Dict
        new_pod_type        # type: Dict
    ):
        # type: (...) -> None
        self._validate_update_pos_values(
            current_pos_values=current_pod_type,
            new_pos_values=new_pod_type,
            key_index="pod_type",
            validate_state=False
        )

    def _delete_directories(self):
        self.updater_service.delete_directory(self.work_path)

    def _check_update_window(self):
        # type: () -> None

        start_time = self.local_configs.window.start_time
        end_time = self.local_configs.window.end_time
        if not ValidationUtil().is_inside_allowed_window(start_time, end_time):
            str_start_time = start_time.strftime("%H:%M:%S")
            str_end_time = end_time.strftime("%H:%M:%S")
            message = "Not inside allowed window. Start: {} / End: {}".format(str_start_time, str_end_time)
            logger.info(message)
            raise NotInsideUpdateWindow()

    def _build_params(self):
        params = {}
        last_versions, _ = self.version_repository.get_last_versions()
        binaries = last_versions[UpdateType.binaries.value] if UpdateType.binaries.value in last_versions else ""
        pos_version = binaries.split("|")[1].split(":")[1] if binaries != "" else ""
        params['pos_version'] = pos_version
        return params

    def _download_new_loaders(self):
        # type: () -> None

        try:
            logger.info("Downloading new loaders")
            file_path = os.path.join(self.downloads_path, self.loaders_zip)
            params = self._build_params()
            self.api_repository.download_loaders(file_path, params)
            logger.info("Loaders successfully downloaded")
        except Exception as ex:
            logger.exception("Error downloading loaders")
            raise ErrorDownloadingLoaders(ex.message)

    def _create_backup(self):
        # type: () -> str

        try:
            logger.info("Creating loaders backup")

            backup_zip_name = "loaders_{}".format(datetime.now().strftime("%Y%m%d_%H%M%S"))
            backup_package = os.path.join(self.backups_path, backup_zip_name)
            backup_package_zip = os.path.join(self.backups_path, backup_zip_name) + ".zip"

            zip_path = os.path.join(self.work_path, self.data_relative_path)
            self.updater_service.copy_directory(self.data_path, zip_path)
            self.updater_service.save_directory_zip(zip_path, backup_package)
            self.updater_service.test_zip_file(backup_package_zip)
            self._clean_backups()

            logger.info("Loaders backup successfully created!")
        except Exception as _:
            logger.exception("Error creating loaders backup")
            raise ErrorCreatingBackup()

        return backup_package_zip

    def _clean_backups(
        self,   # type: LoaderUpdaterService
    ):
        # type: (...) -> None

        list_of_files = os.listdir(self.backups_path)
        if len(list_of_files) > 5:
            full_path_files = [os.path.join(self.backups_path, x) for x in list_of_files]
            full_path_files = sorted(full_path_files, key=lambda p: os.path.getmtime(p))
            files_to_delete = full_path_files[:-5]
            for backup_file in files_to_delete:
                try:
                    os.remove(backup_file)
                except Exception as e:
                    logger.exception("Failed to remove backup file {}. Error {}".format(backup_file, e))

    def _delete_content(self, path):
        # type: (str) -> None

        try:
            logger.info("Deleting path {}".format(path))

            self.updater_service.delete_directory(path)

            logger.info("Path deleted")
        except Exception as _:
            logger.exception("Error deleting path")
            raise ErrorDeletingContent()

    def _get_new_loaders(self):
        # type: () -> None

        try:
            logger.info("Extracting loaders in data path")

            self.updater_service.create_dir(self.data_path)
            zip_path = os.path.join(self.downloads_path, self.loaders_zip)
            self.updater_service.extract_zip(zip_path, self.data_path)

            logger.info("Loaders extracted in data path")
        except Exception as _:
            logger.exception("Error extracting loaders to data path")
            raise ErrorExtractingLoaders()

    def _recover_folders(self):
        # type: () -> None

        current_folder = None
        try:
            work_path_server_folder = os.path.join(self.work_path, self.server_relative_path)
            if not os.path.exists(work_path_server_folder):
                return

            for current_folder in os.listdir(work_path_server_folder):
                if current_folder == "bundles":
                    continue

                logger.info("Recovering {} folder".format(current_folder))
                folder_path = os.path.join(self.server_path, current_folder)
                work_path_path = os.path.join(work_path_server_folder, current_folder)
                if not os.path.isdir(work_path_path):
                    continue

                self.updater_service.copy_directory(work_path_path, folder_path)
                logger.info("{} folder recovered".format(current_folder))
        except Exception as _:
            logger.exception("Error recovering {}".format(current_folder))
            raise ErrorRecoveringFolder()

    def _rollback(self, backup_package_zip):
        # type: (str) -> None

        logger.info("Starting rollback")

        retries = 0
        max_retries = 2

        while True:
            try:
                self._delete_content(self.data_path)
                self.updater_service.extract_zip(backup_package_zip, self.data_path)
            except Exception as _:
                logger.exception("Error while executing rollback")
                if retries == max_retries:
                    raise RollbackError()

                retries += 1
                continue

            logger.info("Rollback finished")
            break

    def _update_ui_customizations(
        self,
    ):
        # (...) -> None

        try:
            customization_kui_dir = os.path.join(
                self.server_path,
                "bundles",
                "customization",
                "kui",
            )
            if not os.path.isdir(customization_kui_dir):
                return

            self.updater_service.create_dir(
                path=self.htdocs_kui_path,
                delete_if_exists=False,
            )

            os_walk = os.walk(customization_kui_dir)
            for step_index, (base_dir, dirs, files) in enumerate(os_walk):
                for dir_name in dirs:
                    file_path = os.path.join(self.htdocs_kui_path, dir_name)
                    self.updater_service.create_dir(
                        path=file_path,
                        delete_if_exists=False,
                    )

                last_path_component = os.path.split(base_dir)[1]
                use_last_path_component = step_index != 0
                for file_name in files:
                    src_file = os.path.join(
                        base_dir,
                        file_name,
                    )
                    if use_last_path_component:
                        dst_file = os.path.join(
                            self.htdocs_kui_path,
                            last_path_component,
                            file_name,
                        )
                    else:
                        dst_file = os.path.join(
                            self.htdocs_kui_path,
                            file_name,
                        )
                    shutil.copy2(
                        src_file,
                        dst_file,
                    )
        except Exception as _:
            logger.exception("Error updating customization files")

    @staticmethod
    def _validate_pos_state(
        model   # type: eTree.XML
    ):
        # type: (...) -> None

        pos_state = model.find("PosState").get("state")
        if pos_state not in ("CLOSED", "UNDEFINED"):
            logger.exception("Error, trying to update loader with day {}".format(pos_state))
            raise ErrorValidatingOpenedUsers("ERRORDAYISOPEN")



# cusip.py - functions for handling CUSIP numbers
#
# Copyright (C) 2015-2022 Arthur de Jong
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

"""CUSIP number (financial security identification number).

CUSIP (Committee on Uniform Securities Identification Procedures) numbers are
used to identify financial securities. CUSIP numbers are a nine-character
alphanumeric code where the first six characters identify the issuer,
followed by two digits that identify and a check digit.

More information:

* https://en.wikipedia.org/wiki/CUSIP
* https://www.cusip.com/

>>> validate('DUS0421C5')
'DUS0421C5'
>>> validate('DUS0421CN')
Traceback (most recent call last):
    ...
InvalidChecksum: ...
>>> to_isin('91324PAE2')
'US91324PAE25'
"""

from stdnum.exceptions import *
from stdnum.util import clean


def compact(number):
    """Convert the number to the minimal representation. This strips the
    number of any valid separators and removes surrounding whitespace."""
    return clean(number, ' ').strip().upper()


_alphabet = '0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ*@#'


def calc_check_digit(number):
    """Calculate the check digits for the number."""
    # convert to numeric first, then sum individual digits
    number = ''.join(
        str((1, 2)[i % 2] * _alphabet.index(n)) for i, n in enumerate(number))
    return str((10 - sum(int(n) for n in number)) % 10)


def validate(number):
    """Check if the number provided is valid. This checks the length and
    check digit."""
    number = compact(number)
    if not all(x in _alphabet for x in number):
        raise InvalidFormat()
    if len(number) != 9:
        raise InvalidLength()
    if calc_check_digit(number[:-1]) != number[-1]:
        raise InvalidChecksum()
    return number


def is_valid(number):
    """Check if the number provided is valid. This checks the length and
    check digit."""
    try:
        return bool(validate(number))
    except ValidationError:
        return False


def to_isin(number):
    """Convert the number to an ISIN."""
    from stdnum import isin
    return isin.from_natid('US', number)



import iso8601
import datetime
import base64

from xml.etree import ElementTree as eTree
from common import TipBuilderCommon
from helper import (
    remove_accents,
    custom_encode_cp860,
)
from old_helper import (
    remove_xml_namespace,
    round_half_away_from_zero,
)
from sysactions import translate_message
from typing import (
    Dict,
    Optional,
    Tuple,
)
from models import LoyaltyConfigurations


class PrinterCommon(object):
    def __init__(
        self,
        tip_config,                 # type: Dict
        tip_config_range,           # type: Dict
        tip_builder_common,         # type: TipBuilderCommon
        customer_call_by_field,     # type: str
    ):
        # type: (...) -> None

        self.tip_config = tip_config
        self.tip_config_range = tip_config_range
        self.tip_builder_common = tip_builder_common
        self.customer_call_by_field = customer_call_by_field

    def get_tip_order(
        self,
        order,              # type: eTree
        is_mobile=False     # type: bool
    ):
        # type: (...) -> str

        if len(self.tip_config_range) == 0 and len(self.tip_config) == 0:
            return ""

        order_tip = float(order.get("tip"))
        if len(self.tip_config_range) == 0 and len(self.tip_config) > 0:
            return self._get_tip_line(self.tip_config['TipCode'], self.tip_config['TipName'], order_tip, is_mobile)

        report = ""

        total_gross = float(order.get('totalGross'))
        tip_ranges = self.tip_builder_common.get_percentage_by_tip(total_gross, order_tip)
        for key, value in tip_ranges.items():
            report += self._get_tip_line(key, value.get("TipName"), value.get("TipValue"), is_mobile)

        return report

    def get_coupon_call_text(
        self,
        model,  # type: eTree.Element
    ):
        # type: (...) -> str

        if self.customer_call_by_field:
            call_by = custom_encode_cp860(
                text=self.customer_call_by_field,
            )
            coupon_call_text = "{}".format(call_by)
        else:
            panel_call_text = custom_encode_cp860(
                text=translate_message(
                    model=model,
                    message_key="CALL_PANEL",
                ),
            )
            will_be_called_text = custom_encode_cp860(
                text=translate_message(
                    model=model,
                    message_key="YOU_WILL_BE_CALLED",
                ),
            )

            coupon_call_text = "{} - {}".format(
                panel_call_text,
                will_be_called_text,
            )

        return coupon_call_text

    @staticmethod
    def value_tip_by_percentage(percent, order_value):
        # type: (float, float) -> float

        return (percent * order_value) / 100.0

    @staticmethod
    def _get_tip_line(
        code,           # type: str
        description,    # type: str
        value,          # type: float
        is_mobile       # type: bool
    ):
        # type: (...) -> str

        format_mobile = "%-7.7s|%-14.14s|%6.3f|%-3.2s|%9.2f|"
        format_desktop = "%-7.7s %-18.18s%6.3f%2.2s%9.02f"
        order_tip_format = format_mobile if is_mobile else format_desktop

        tip_line = remove_accents(order_tip_format % (code, description, float(1), "un", value))

        format_mobile = "%-39.40s|%7.02f\n"
        format_desktop = "%-46.46s %9.02f\n"
        order_tip_format = format_mobile if is_mobile else format_desktop

        return order_tip_format % (tip_line, value)

    @staticmethod
    def get_coupon_call_id(
        order,                      # type: eTree.Element
        order_ref_ids,              # type: list
        customer_name_limit=None,   # type: Optional[int]
    ):
        # type: (...) -> Optional[str]

        if order_ref_ids[0] == "HIDE":
            return None

        pager_number = order.find("CustomOrderProperties/OrderProperty[@key='PAGER_NUMBER']")
        if pager_number is not None:
            pager_number = pager_number.get("value")

        customer_name = order.find("CustomOrderProperties/OrderProperty[@key='CUSTOMER_NAME']")
        if customer_name is not None:
            customer_name = customer_name.get("value")

        for ref in order_ref_ids:
            if pager_number and ref == "PAGER_NUMBER":
                return pager_number
            elif customer_name and ref == "CUSTOMER_NAME":
                if customer_name_limit:
                    customer_name = customer_name[:customer_name_limit]
                return custom_encode_cp860(text=customer_name)
            elif ref == "ORDER_ID":
                return order.get('orderId')[-3:]

        return None

    @staticmethod
    def get_sat_params_to_print_loyalty_coupon(
        fiscal_request,     # type: eTree.Element
        order,              # type: eTree.Element
    ):
        # type: (...) -> Tuple[datetime, str]

        fiscal_request_xml = eTree.XML(base64.b64decode(fiscal_request.text.strip()))
        date = fiscal_request_xml.find(".//dEmi").text
        hour = fiscal_request_xml.find(".//hEmi").text
        protocol_date = datetime.datetime.strptime(
            "{}{}".format(
                date,
                hour,
            ),
            "%Y%m%d%H%M%S",
        )
        sat_id = fiscal_request_xml.find(".//infCFe").get("Id")[3:]
        qr_code_text = "{}|{:.2f}".format(
            sat_id,
            round_half_away_from_zero(
                number=order.get("totalTender"),
                precision=2,
            )
        )

        return protocol_date, qr_code_text

    @staticmethod
    def get_nfce_params_to_print_loyalty_coupon(
        fiscal_request,     # type: eTree.Element
        order,              # type: eTree.Element
    ):
        # type: (...) -> Tuple[datetime, str]

        fiscal_request_xml = remove_xml_namespace(
            eTree.tostring(
                element=fiscal_request,
                encoding='utf8',
                method='xml',
            ),
        )
        protocol_date = iso8601.parse_date(fiscal_request_xml.find(".//dhEmi").text)
        nfce_id = fiscal_request_xml.find(".//infNFe").get("Id")[3:]
        qr_code_text = "{}|{:.2f}".format(
            nfce_id,
            round_half_away_from_zero(
                number=order.get("totalTender"),
                precision=2,
            )
        )

        return protocol_date, qr_code_text

    @staticmethod
    def should_print_loyalty_coupon(
        order,                   # type: eTree.Element
        loyalty_configurations,  # type: LoyaltyConfigurations
    ):
        # type: (...) -> bool

        if not loyalty_configurations.print_loyalty_coupon_after_fiscalization:
            return False

        has_loyalty_id = order.find("CustomOrderProperties/OrderProperty[@key='LOYALTY_ID']") is not None
        if not has_loyalty_id:
            return True

        return False



# -*- coding: utf-8 -*-

import base64
import json
import os
import time

import re

from modelcommon.model import SitefResponse
from pos_model import TenderType

from typing import (
    Any,
    Union,
    Optional,
)

from model import (
    CancelPaymentOrderData,
    ConfirmFinishPaymentData,
    FinishPaymentData,
    PaymentResponseMessage,
    PaymentResponseData,
    TestConnectionData,
    StartPaymentData,
    PaymentResponse,
    TPag,
)
from modelcommon.electronicpayment import TransactionProcessors
from paymentprocessor import PaymentProcessor
from _CapptaCallback import CapptaCallback
from util import get_transaction_identifier

cappta_client = None


class CapptaProcessor(PaymentProcessor):

    def __init__(
        self,
        message_bus,        # type: MbContextMessageBus
        dialog_manager,     # type: DialogManager
        config,             # type: Group
        logger,             # type: Logger
        cache,              # type: Cache
    ):
        # type: (...) -> None

        super(CapptaProcessor, self).__init__(message_bus, dialog_manager, config, logger, cache)
        self.cnpj = config.find_value("CnpjLoja")
        self.pos_id = config.find_value("PosId")
        self.pdv_tef = config.find_value("PdvTef")
        self.auth_key = config.find_value("ChaveAutenticacao")
        self.administrative_password = config.find_value("AdministrativePassword", "cappta")
        self.callback = CapptaCallback(self.dialog_manager)
        self.DetalhesCredito = None
        self.IMensagem = None
        self.IRespostaTransacaoPendente = None
        self.IRequisicaoParametro = None
        self.IRespostaOperacaoAprovada = None
        self.IRespostaOperacaoRecusada = None
        self.cappta_client = self._get_cappta_client(config)
        self.cache = cache

    def connection_status(
        self,
        test_connection_data    # type: TestConnectionData
    ):
        # type: (...) -> bool

        return True

    def add_payment(
        self,
        start_payment_data,     # type: StartPaymentData
    ):
        # type: (...) -> PaymentResponse

        try:
            transaction_identifier = get_transaction_identifier(
                order_id=start_payment_data.order_id,
                tender_id=start_payment_data.tender_id,
            )
            ret = self._authenticate()
            if ret == 0:
                self.cappta_client.IniciarMultiCartoes(9)
                tender_type = start_payment_data.electronic_type

                amount = float(start_payment_data.amount.replace(".", "").replace(",", "."))

                if tender_type == TenderType.credit:
                    ret = self.cappta_client.PagamentoCredito(amount, self.DetalhesCredito())
                elif tender_type in (TenderType.debit, TenderType.voucher):
                    ret = self.cappta_client.PagamentoDebito(amount)

                if ret != 0:
                    self.callback.callback(str(self.pos_id), ret, "")
                    return PaymentResponseMessage("Pagamento interrompido")

                tef_loop_variable = self.cappta_client.IterarOperacaoTef()
                while True:
                    if isinstance(tef_loop_variable, self.IMensagem):
                        self._execute_message_response(
                            tef_loop_variable=tef_loop_variable,
                        )

                    elif isinstance(tef_loop_variable, self.IRespostaTransacaoPendente):
                        ret = self._execute_pending_transaction_response(
                            tef_loop_variable=tef_loop_variable,
                        )

                    elif isinstance(tef_loop_variable, self.IRequisicaoParametro):
                        ret = self._execute_request_parameter(
                            tef_loop_variable=tef_loop_variable,
                        )

                    is_approved_operation = isinstance(tef_loop_variable, self.IRespostaOperacaoAprovada)
                    is_denied_operation = isinstance(tef_loop_variable, self.IRespostaOperacaoRecusada)
                    if is_approved_operation or is_denied_operation:
                        break

                    tef_loop_variable = self.cappta_client.IterarOperacaoTef()

                if isinstance(tef_loop_variable, self.IRespostaOperacaoAprovada):
                    approved_payment = self.IRespostaOperacaoAprovada(tef_loop_variable)
                    response = self._get_payment_response(
                        approved_payment=approved_payment,
                        start_payment_data=start_payment_data,
                        tender_type=tender_type,
                    )
                    response.result = ret
                    response.approved_amount = start_payment_data.amount
                    response.electronic_type = start_payment_data.electronic_type
                    response.transaction_identifier = transaction_identifier

                    msg = "POS %s - PaymentApproved. Order %s" % (str(self.pos_id), str(start_payment_data.order_id))
                    self.logger.debug(msg)
                    return response

                elif isinstance(tef_loop_variable, self.IRespostaOperacaoRecusada):
                    response = self._execute_action_operation_declined(
                        order_id=start_payment_data.order_id,
                        tef_loop_variable=tef_loop_variable,
                    )
                    return PaymentResponseMessage(response)
            else:
                message = "Falha ao autenticar POS %s - Código de erro %s" % (str(self.pos_id), str(ret))
                self.logger.error(message)
                return PaymentResponseMessage(message)
        finally:
            self.dialog_manager.close_dialogs(start_payment_data.pos_id)

    def _get_payment_response(
        self,
        approved_payment,       # type: IRespostaOperacaoAprovada
        start_payment_data,     # type: StartPaymentData
        tender_type,            # type: TenderType
    ):
        # type: (...) -> PaymentResponseData

        customer_slip = approved_payment.CupomLojista
        b64_receipt_merchant = base64.b64encode(str(customer_slip).replace("\"", ""))
        client_slip = approved_payment.CupomCliente
        b64_receipt_customer = base64.b64encode(str(client_slip).replace("\"", ""))
        self._set_cache_receipt_data(
            order_id=start_payment_data.order_id,
            tender_id=start_payment_data.tender_id,
            receipt_customer=b64_receipt_customer,
            receipt_merchant=b64_receipt_merchant,
        )

        return PaymentResponseData(
            media=str(approved_payment.NomeBandeiraCartao),
            result="",
            card_number=str(approved_payment.NumeroCartaoCliente)[:6],
            auth_code=str(approved_payment.CodigoAutorizacaoAdquirente),
            approved_amount="",
            id_auth=approved_payment.NumeroControle,
            cnpj_auth="",
            bandeira=str(approved_payment.CodigoBandeiraCartao),
            receipt_merchant=b64_receipt_merchant,
            receipt_customer=b64_receipt_customer,
            expiration_date="",
            nsu=approved_payment.NsuTef,
            transaction_processor=TransactionProcessors.CAPPTA,
            last_digits=str(approved_payment.NumeroCartaoCliente)[4:],
            sat_acquirer="",
            t_pag=TPag.CREDITO if tender_type == TenderType.credit else TPag.DEBITO,
            t_band=str(approved_payment.CodigoBandeiraCartao),
            nsu_host=approved_payment.NsuAdquirente,
            electronic_type="",
            transaction_identifier="",
        )

    def cancel_payment_order(
        self,
        cancel_payment_order_data,  # type: CancelPaymentOrderData
        additional_params="",       # type: Optional[str]
    ):
        # type: (...) -> Any[Union[SitefResponse, str]]

        try:
            return self.execute_cancel_payment_order(
                cancel_payment_order_data=cancel_payment_order_data,
            )
        finally:
            pos_id = cancel_payment_order_data.pos_id
            self.dialog_manager.close_dialogs(
                pos_id=pos_id,
            )

    def execute_cancel_payment_order(
        self,
        cancel_payment_order_data    # type: CancelPaymentOrderData
    ):
        # type: (...) -> Any[Union[SitefResponse, str]]

        self.logger.debug("CapptaProcessor Cancel payment order - STARTED")

        try:
            transaction_identifier = get_transaction_identifier(
                order_id=cancel_payment_order_data.order_id,
                tender_id=cancel_payment_order_data.tender_id,
            )
            result = self._authenticate()
            if result == 0:

                result = self.cappta_client.CancelarPagamento(
                    self.administrative_password,
                    cancel_payment_order_data.id_auth,
                )

                if result == 0:
                    tef_loop_variable = self.cappta_client.IterarOperacaoTef()
                    while True:
                        if isinstance(tef_loop_variable, self.IMensagem):
                            self._execute_message_response(
                                tef_loop_variable=tef_loop_variable,
                                card_number=cancel_payment_order_data.card_number,
                            )

                        elif isinstance(tef_loop_variable, self.IRespostaTransacaoPendente):
                            self._execute_pending_transaction_response(
                                tef_loop_variable=tef_loop_variable,
                            )

                        elif isinstance(tef_loop_variable, self.IRequisicaoParametro):
                            self._execute_request_parameter(
                                tef_loop_variable=tef_loop_variable,
                            )

                        is_approved_operation = isinstance(tef_loop_variable, self.IRespostaOperacaoAprovada)
                        is_denied_operation = isinstance(tef_loop_variable, self.IRespostaOperacaoRecusada)
                        if is_approved_operation or is_denied_operation:
                            break

                        tef_loop_variable = self.cappta_client.IterarOperacaoTef()

                    if isinstance(tef_loop_variable, self.IRespostaOperacaoAprovada):
                        approved_cancel = self.IRespostaOperacaoAprovada(tef_loop_variable)
                        response = self._get_cancel_response(
                            approved_cancel=approved_cancel,
                            cancel_payment_order_data=cancel_payment_order_data,
                        )
                        response.transaction_identifier = transaction_identifier
                        response.transaction_value = cancel_payment_order_data.tender_amount

                        return response

                    elif isinstance(tef_loop_variable, self.IRespostaOperacaoRecusada):
                        response = self._execute_action_operation_declined(
                            order_id=cancel_payment_order_data.order_id,
                            tef_loop_variable=tef_loop_variable,
                        )
                        return PaymentResponseMessage(response)

            return PaymentResponseMessage("$ERROR_CANCEL_ELECTRONIC_PAYMENTS_ORDER")
        except Exception as _:
            self.logger.error("CapptaProcessor Cancel payment order - ERROR")
            return PaymentResponseMessage("$ERROR_CANCEL_ELECTRONIC_PAYMENTS_ORDER")
        finally:
            self.logger.debug("CapptaProcessor Cancel payment order - FINISH")

    def _get_cancel_response(
        self,
        approved_cancel,                # type: IRespostaOperacaoAprovada
        cancel_payment_order_data,      # type: CancelPaymentOrderData
    ):
        # type: (...) -> SitefResponse

        customer_slip = approved_cancel.CupomLojista
        b64_receipt_merchant = base64.b64encode(str(customer_slip).replace("\"", ""))
        client_slip = approved_cancel.CupomCliente
        b64_receipt_customer = base64.b64encode(str(client_slip).replace("\"", ""))
        self._set_cache_receipt_data(
            order_id=cancel_payment_order_data.order_id,
            tender_id=cancel_payment_order_data.tender_id,
            receipt_customer=b64_receipt_customer,
            receipt_merchant=b64_receipt_merchant,
        )
        response = SitefResponse()
        response.customer_receipt = b64_receipt_customer
        response.merchant_receipt = b64_receipt_merchant
        response.nsu = approved_cancel.NsuAdquirente
        response.nsu_host = approved_cancel.NsuTef

        return response

    def _execute_message_response(
        self,
        tef_loop_variable,       # type: IMensagem
        card_number=None,        # type: str
    ):
        # type: (...) -> None

        message = self.IMensagem(tef_loop_variable)
        msg = message.Descricao.encode('utf-8')

        if card_number is not None and 'insira' in msg.lower():
            msg = "$INSERT_CARD_CANCEL_PAYMENT|{}|".format(card_number)

        self.callback.callback(
            pos_id=str(self.pos_id),
            message=msg,
            callback_type="MENSAGEM",
        )
        time.sleep(0.5)

    def _execute_pending_transaction_response(
        self,
        tef_loop_variable,      # type: IRespostaTransacaoPendente
    ):
        # type: (...) -> int

        pending_transactions = self.IRespostaTransacaoPendente(tef_loop_variable)
        msg = pending_transactions.Mensagem.encode('utf-8')
        callback_response = self.callback.callback(
            pos_id=str(self.pos_id),
            message=msg,
            callback_type="REQUISITA_ACAO",
        )
        if callback_response == "CANCELA":
            callback_response = "2"

        return self.cappta_client.EnviarParametro(callback_response, 1)

    def _execute_request_parameter(
        self,
        tef_loop_variable,      # type: IRequisicaoParametro
    ):
        # type: (...) -> int

        request_parameter = self.IRequisicaoParametro(tef_loop_variable)
        request_message = request_parameter.Mensagem.encode('utf-8')

        if 'digite o cartao' in request_message.lower():
            self.dialog_manager.show_message(
                pos_id=str(self.pos_id),
                message="",
            )
            return self.cappta_client.EnviarParametro("", 2)

        if "para cancelar" in request_message.lower():
            request_message = request_message.split("\n")
            request_message = request_message[len(request_message) - 1]
            callback_response = re.findall('[0-9]+', request_message)
            if callback_response:
                return self.cappta_client.EnviarParametro(str(callback_response[0]), 1)

        if '(1-Sim 0-Não)' in request_message:
            request_message = request_message.replace("(1-Sim 0-Não)", "")

        callback_response = self.dialog_manager.show_question(
            pos_id=str(self.pos_id),
            message=request_message,
        )

        callback_response = "1" if callback_response == "0" else "0"

        return self.cappta_client.EnviarParametro(str(callback_response), 1)

    def _execute_action_operation_declined(
        self,
        order_id,           # type: str
        tef_loop_variable,  # type: IRespostaOperacaoRecusada
    ):

        # type: (...) -> str

        operation = self.IRespostaOperacaoRecusada(tef_loop_variable)
        response = operation.Motivo.encode('utf-8')
        log = "POS {0} - Operação recusada. Order: {1}. Motivo: {2}".format(self.pos_id, order_id, response)
        self.logger.debug(log)
        self.callback.callback(
            pos_id=str(self.pos_id),
            message=response,
            callback_type="MENSAGEM",
        )
        time.sleep(0.5)

        return response

    def cancel_payment(self):
        return

    def finish_payment(
        self,
        finish_payment_data,     # type: FinishPaymentData
    ):
        # type: (...) -> bool

        try:
            if finish_payment_data.status == 1:
                self.cappta_client.ConfirmarPagamentos()
            else:
                self.cappta_client.DesfazerPagamentos()
            return True
        except Exception as _:
            return False

    def confirm_finish_payment(
        self,
        confirm_finish_payment_data     # type: ConfirmFinishPaymentData
    ):
        # type: (...) -> None

        pass

    def get_name(
        self,
    ):
        # type: (...) -> str

        self.register_eft_processor(
            pos_id=self.pos_id,
            processor_name=TransactionProcessors.CAPPTA,
        )

        return TransactionProcessors.CAPPTA

    def get_document(
        self,
        pos_id      # type: str
    ):
        # type: (...) -> str

        pass

    def stop(self):
        return

    def get_last_receipt(
        self,
        eft_processor,  # type: str
    ):
        # type: (...) -> Optional[str]

        return self.cache.get_last_receipt(eft_processor=TransactionProcessors.CAPPTA)

    def _set_cache_receipt_data(
        self,
        order_id,           # type: str
        tender_id,          # type: str
        receipt_customer,   # type: str
        receipt_merchant,   # type: str
    ):
        # type: (...) -> None

        receipt_data = {
            "order_id": order_id,
            "tender_id": tender_id,
            "receipt_customer": receipt_customer,
            "receipt_merchant": receipt_merchant,
        }
        receipt_data = json.dumps(receipt_data)
        self.cache.set_last_receipt(
            receipt_data=receipt_data,
            eft_processor=TransactionProcessors.CAPPTA,
        )

    def _get_cappta_client(self, config):
        import clr  # Linux (POS Mobile) does not have an implemented library to clr

        global cappta_client
        if cappta_client is not None:
            return cappta_client

        dll_path = self._get_dll_path(config.find_value("DLLPath"))
        old_dir = os.getcwd()
        os.chdir(dll_path)
        clr.AddReference(os.path.join(os.getcwd(), "Cappta.Gp.Api.Com.dll"))

        from Cappta.Gp.Api.Com import ClienteCappta

        cappta_client = ClienteCappta()
        os.chdir(old_dir)

        return cappta_client

    @staticmethod
    def _get_dll_path(dll_path):
        for var in re.findall("{[a-zA-Z0-9]*}", dll_path):
            dll_path = dll_path.replace(var, os.environ.get(var.replace('{', '').replace('}', ''), ""))
        return os.path.normpath(dll_path)

    def _authenticate(
        self
    ):
        # type: (...) -> int

        result = self.cappta_client.AutenticarPdv(
            self.cnpj,
            self.pdv_tef,
            self.auth_key,
        )

        if result != 0:
            self.callback.callback(
                pos_id=str(self.pos_id),
                message="$AUTHENTICATION_FAILED",
                callback_type="MENSAGEM",
            )
            time.sleep(3)
            return result

        self.logger.debug("POS %s - Autenticação com sucesso." % (str(self.pos_id)))

        from Cappta.Gp.Api.Com.Model import (
            Configuracoes,
            DetalhesCredito,
            IMensagem,
            IRespostaTransacaoPendente,
            IRequisicaoParametro,
            IRespostaOperacaoAprovada,
            IRespostaOperacaoRecusada,
        )

        self.DetalhesCredito = DetalhesCredito
        self.IMensagem = IMensagem
        self.IRespostaTransacaoPendente = IRespostaTransacaoPendente
        self.IRequisicaoParametro = IRequisicaoParametro
        self.IRespostaOperacaoAprovada = IRespostaOperacaoAprovada
        self.IRespostaOperacaoRecusada = IRespostaOperacaoRecusada

        configs = Configuracoes()
        configs.ExibirInterface = False
        result_config = self.cappta_client.Configurar(configs)
        if result_config != 0:
            self.logger.debug("Não foi possível realizar a configuração")

        return result



# -*- coding: utf-8 -*-
from cache_functions import UsersInfoCache
from logging import Logger

from messageprocessorutil import LoggingProcessorCallback
from messagebus import TokenCreator, TokenPriority, MessageBus
from messagehandler import MessageHandlerBuilder
from messageprocessor import (
    DefaultMessageProcessorExecutorFactory,
    MessageProcessorMessageHandler,
)
from production.box import ViewBox

from production.manager import ProductionManager, PicklistCleanerManager
from production.interactor import DeviceStatusChangedInteractor
from production.processor import (
    OrderModifiedProcessor,
    GetOrderMaxPrepTimeProcessor,
    GetOrderXmlProcessor,
    OrderModifiedIgnoredTypesProcessor,
    StatisticUpdateProcessor,
    PurgeOrdersProcessor,
    GetPrintersInfoProcessor,
    PickListOrdersProcessor,
    DeviceStatusChangedProcessor,
    GetCoursesProcessor,
)
from production.processor.command import SetOrderStateProcessor, ToggleTagLineProcessor, UndoProcessor, \
    RefreshViewProcessor, EnablePathProcessor, DisablePathProcessor
from messagehandlerbuilder.token import ListenedEvents
from production.repository import ProductRepository, ProductionRepository
from production.treebuilder import TreeBuilder
from production.treebuilder.boxfactory import ProductionBoxFactory
from production.treebuilder.viewfactory import ViewFactory
from production.util import ProductionConfig
from typing import List, Dict, Union

production_group = "C"
TK_PROD_REFRESH_VIEW = TokenCreator.create_token(TokenPriority.low, production_group, "1")
TK_PROD_SET_ORDER_STATE = TokenCreator.create_token(TokenPriority.low, production_group, "2")
TK_PROD_GET_ORDER_MAX_PREP_TIME = TokenCreator.create_token(TokenPriority.low, production_group, "3")
TK_PROD_GET_ORDER_XML = TokenCreator.create_token(TokenPriority.low, production_group, "4")  # 12582916
TK_PROD_UNDO = TokenCreator.create_token(TokenPriority.low, production_group, "7")
TK_PROD_TOGGLE_TAG_LINE = TokenCreator.create_token(TokenPriority.low, production_group, "8")

TK_PROD_DISABLE_PATH = TokenCreator.create_token(TokenPriority.low, production_group, "101")
TK_PROD_ENABLE_PATH = TokenCreator.create_token(TokenPriority.low, production_group, "100")
TK_PROD_GET_STATISTICS = TokenCreator.create_token(TokenPriority.low, production_group, "102")
TK_PROD_PURGE = TokenCreator.create_token(TokenPriority.low, production_group, "103")

TK_PROD_GET_PRINTERS_INFO = TokenCreator.create_token(TokenPriority.low, production_group, "201")
TK_PROD_GET_COURSES = TokenCreator.create_token(TokenPriority.low, production_group, "202")

TK_PROD_ORDER_PICKLIST = TokenCreator.create_token(TokenPriority.low, production_group, "301")  # 12583681

TK_SALE_ORDER_MODIFIED = TokenCreator.create_token(TokenPriority.low, "A", "32")


class ProductionMessageHandlerBuilder(MessageHandlerBuilder):
    def __init__(
        self,                          # type: ProductionMessageHandlerBuilder
        message_bus,                   # type: MessageBus
        purge_interval,                # type: int
        orders_life_time,              # type: int
        product_repository,            # type: ProductRepository
        production_repository,         # type: ProductionRepository
        manager_log_level,             # type: str
        audit_logger_prod_states,      # type: List
        kds_manager,                   # type: str
        order_modified_by_message,     # type: bool
        production_config,             # type: ProductionConfig
        loggers,                       # type: Dict[unicode, Logger]
        timeout_to_cleanup_devices,    # type: int
    ):
        # type: (...) -> None

        self.message_bus = message_bus
        self.purge_interval = purge_interval
        self.orders_life_time = orders_life_time
        self.product_repository = product_repository
        self.production_repository = production_repository
        self.manager_log_level = manager_log_level
        self.audit_logger_prod_states = audit_logger_prod_states
        self.production_config = production_config
        self.boxes_config = production_config.boxes_config
        self.branches_config = production_config.branches_config
        self.views_config = production_config.views_config
        self.clean_picklist_database_service_config = production_config.picklist_database_manager_config
        self.kds_manager = kds_manager
        self.order_modified_by_message = order_modified_by_message
        self.loggers = loggers
        self.timeout_to_cleanup_devices = timeout_to_cleanup_devices

        self.production_manager = None  # type: Union[None, ProductionManager]
        self.message_handler = None
        self.user_cache_thread = None
        self.clean_picklist_database_service = None

    def build_singletons(self):
        if self.manager_log_level in self.loggers:
            manager_logger = self.loggers[self.manager_log_level]
        elif "ERROR" in self.loggers:
            manager_logger = self.loggers["ERROR"]
        elif len(self.loggers) > 0:
            manager_logger = self.loggers[self.loggers.keys()[0]]
        else:
            manager_logger = None

        self.production_manager = ProductionManager(
            self.production_repository,
            self.purge_interval,
            self.orders_life_time,
            self.message_bus,
            self.audit_logger_prod_states,
            self.kds_manager,
            manager_logger
        )

        production_box_factory = ProductionBoxFactory(
            self.production_manager,
            self.production_manager,
            self.production_manager,
            self.production_manager,
            self.product_repository,
            self.loggers
        )

        view_factory = ViewFactory(
            message_bus=self.message_bus,
            loggers=self.loggers,
            production_manager=self.production_manager,
            kds_manager=self.kds_manager,
            timeout_to_cleanup_devices=self.timeout_to_cleanup_devices,
        )
        tree_builder = TreeBuilder(production_box_factory, view_factory)
        trees = tree_builder.build_trees(
            self.boxes_config,
            self.branches_config,
            self.views_config
        )

        root_boxes = []
        view_boxes = []
        for tree in trees:
            root_boxes.append(tree.root)
            leaves = [t for t in tree.leaves if isinstance(t, ViewBox)]
            view_boxes.extend(leaves)

        self.production_manager.set_root_boxes(root_boxes)
        self.production_manager.set_views(view_boxes)
        self.production_manager.initialize()

        self.user_cache_thread = UsersInfoCache(self.message_bus, manager_logger)
        self.user_cache_thread.daemon = True
        self.user_cache_thread.start()

        self.clean_picklist_database_service = PicklistCleanerManager(
            production_repository=self.production_repository,
            config=self.clean_picklist_database_service_config,
        )
        self.clean_picklist_database_service.start()

        order_modified_processor = OrderModifiedProcessor(
            self.production_manager,
            self.product_repository,
            self.user_cache_thread,
            self.order_modified_by_message
        )

        device_status_interactor = DeviceStatusChangedInteractor(production_manager=self.production_manager)
        device_status_changed_processor = DeviceStatusChangedProcessor(
            device_status_interactor=device_status_interactor,
        )

        event_processors = {
            ListenedEvents.DEVICE_STATUS_CHANGED: device_status_changed_processor,
        }
        if not self.order_modified_by_message:
            order_modified_ignored_types_processor = OrderModifiedIgnoredTypesProcessor()
            event_processors.update({
                ListenedEvents.ORDER_MODIFIED: order_modified_processor,
                ListenedEvents.ORDER_MODIFIED_DO_TENDER: order_modified_ignored_types_processor,
                ListenedEvents.ORDER_MODIFIED_DISCOUNT_APPLIED: order_modified_ignored_types_processor,
                ListenedEvents.ORDER_MODIFIED_DISCOUNT_REMOVED: order_modified_ignored_types_processor,
                ListenedEvents.ORDER_MODIFIED_CLEAR_TENDERS: order_modified_ignored_types_processor,
                ListenedEvents.ORDER_MODIFIED_RESET_CURRENT_ORDER: order_modified_ignored_types_processor,
            })

        max_prep_time_processor = GetOrderMaxPrepTimeProcessor(self.production_manager, self.product_repository)
        message_processors = {
            TK_PROD_SET_ORDER_STATE: SetOrderStateProcessor(view_boxes),
            TK_PROD_TOGGLE_TAG_LINE: ToggleTagLineProcessor(view_boxes, self.loggers),
            TK_PROD_UNDO: UndoProcessor(view_boxes),
            TK_PROD_REFRESH_VIEW: RefreshViewProcessor(view_boxes),
            TK_PROD_ENABLE_PATH: EnablePathProcessor(self.production_manager),
            TK_PROD_DISABLE_PATH: DisablePathProcessor(self.production_manager),
            TK_PROD_GET_ORDER_MAX_PREP_TIME: max_prep_time_processor,
            TK_PROD_GET_ORDER_XML: GetOrderXmlProcessor(self.production_manager),
            TK_PROD_GET_STATISTICS: StatisticUpdateProcessor(self.production_manager),
            TK_PROD_PURGE: PurgeOrdersProcessor(self.production_manager),
            TK_PROD_GET_PRINTERS_INFO: GetPrintersInfoProcessor(view_boxes),
            TK_PROD_GET_COURSES: GetCoursesProcessor(self.product_repository),
            TK_PROD_ORDER_PICKLIST: PickListOrdersProcessor(
                message_bus=self.message_bus,
                production_repository=self.production_repository,
            )
        }

        if not self.order_modified_by_message:
            order_modified_processor = OrderModifiedIgnoredTypesProcessor(True)
        message_processors[TK_SALE_ORDER_MODIFIED] = order_modified_processor

        callbacks = [LoggingProcessorCallback(manager_logger)]

        self.message_handler = MessageProcessorMessageHandler(
            event_processors=event_processors,
            message_processors=message_processors,
            executor_factory=DefaultMessageProcessorExecutorFactory(callbacks),
        )

    def destroy_message_handler(
        self,
        message_handler,    # type: ProductionMessageHandlerBuilder
    ):
        # type: (...) -> None

        pass

    def build_message_handler(self, message, event):
        return self.message_handler

    def destroy_singletons(self):
        self.production_manager.stop()
        self.user_cache_thread.stop()
        self.production_repository.stop()
        self.clean_picklist_database_service.stop()



# -*- coding: utf-8 -*-

class ProcInutNfeBuilder(object):
    @staticmethod
    def build_proc_inut_nfe_xml(request, response, versao_ws):
        # type: (str, str, int) -> str
        index = request.index("<inutNFe ")
        index2 = request.index("</inutNFe>")
        inut_nfe = request[index:index2 + 10]

        index = response.index("<retInutNFe ")
        index2 = response.index("</retInutNFe>")
        ret_inut_nfe = response[index:index2 + 13]

        proc_inut_message = "<ProcInutNFe  xmlns=\"http://www.portalfiscal.inf.br/nfe\" versao=\"%.2f\">"
        proc_inut_nfe = proc_inut_message % (3.1 if versao_ws in (1, 3) else 4)
        proc_inut_nfe += inut_nfe + ret_inut_nfe
        proc_inut_nfe += "</ProcInutNFe>"

        return proc_inut_nfe.encode("utf8")



# -*- coding: utf-8 -*-
from typing import Optional

from nfcebuilder.nfceutil import XmlEnveloper, NfeSigner
from nfcedisabler import NfceDisablerParameters


class NfceDisablerRequestBuilder(object):
    NAMESPACE_NFE = "http://www.portalfiscal.inf.br/nfe"

    def __init__(self, disabler_parameters, nfce_signer, xml_enveloper, versao_ws):
        # type: (NfceDisablerParameters, NfeSigner, XmlEnveloper, int) -> None
        self.disabler_parameters = disabler_parameters
        self.nfce_signer = nfce_signer
        self.xml_enveloper = xml_enveloper
        self.versao_ws = versao_ws
        self.soap_action = None

        if versao_ws == 1:
            self.namespace_ret_inutilizacao = "http://www.portalfiscal.inf.br/nfe/wsdl/NfeInutilizacao2"
        elif versao_ws == 3:
            self.namespace_ret_inutilizacao = "http://www.portalfiscal.inf.br/nfe/wsdl/NfeInutilizacao3"
        else:
            self.namespace_ret_inutilizacao = "http://www.portalfiscal.inf.br/nfe/wsdl/NFeInutilizacao4"

    def build_request(self, order):
        ano_nota = order.get("createdAt")[2:4]
        order_id = order.get("orderId")
        fiscal_id = None  # type: Optional[int]

        fiscal_id_node = order.find('.//OrderProperty[@key="FISCAL_ID"]')
        if fiscal_id_node is not None:
            fiscal_id = int(fiscal_id_node.get("value"))

        parameters_xml = self.disabler_parameters.fill_parameters(ano_nota, order_id, fiscal_id)

        inut_pattern = "<inutNFe xmlns=\"{0:s}\" versao=\"{1:.2f}\">{2:s}</inutNFe>"

        xml = inut_pattern.format(self.NAMESPACE_NFE, 3.1 if self.versao_ws in (1, 3) else 4, parameters_xml)

        signed_xml = self.nfce_signer.sign_xml(xml, "infInut", "Id")

        envelope = self.xml_enveloper.envelop(signed_xml, self.namespace_ret_inutilizacao, self.versao_ws)

        return envelope



from get_nf_type import get_nf_type
from check_certificate import check_certificate
from get_error_fiscal_message import get_error_fiscal_message
from check_can_fiscalize_order_action import check_can_fiscalize_order_action



from datetime import datetime

from typing import Optional


class DeviceStatusDto(object):
    def __init__(
        self,
        type,                       # type: str
        name,                       # type: str
        enabled,                    # type: bool
        last_update=None,           # type: Optional[datetime]
        last_update_operator=None,  # type: Optional[str]
    ):
        # type: (...) -> None

        self.type = type
        self.name = name
        self.enabled = enabled
        self.last_update = last_update
        self.last_update_operator = last_update_operator



# Copyright (C) Dnspython Contributors, see LICENSE for text of ISC license

# Copyright (C) 2003-2017 Nominum, Inc.
#
# Permission to use, copy, modify, and distribute this software and its
# documentation for any purpose with or without fee is hereby granted,
# provided that the above copyright notice and this permission notice
# appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND NOMINUM DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL NOMINUM BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT
# OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

"""DNS TTL conversion."""

import dns.exception
from ._compat import long


class BadTTL(dns.exception.SyntaxError):
    """DNS TTL value is not well-formed."""


def from_text(text):
    """Convert the text form of a TTL to an integer.

    The BIND 8 units syntax for TTLs (e.g. '1w6d4h3m10s') is supported.

    *text*, a ``text``, the textual TTL.

    Raises ``dns.ttl.BadTTL`` if the TTL is not well-formed.

    Returns an ``int``.
    """

    if text.isdigit():
        total = long(text)
    else:
        if not text[0].isdigit():
            raise BadTTL
        total = long(0)
        current = long(0)
        for c in text:
            if c.isdigit():
                current *= 10
                current += long(c)
            else:
                c = c.lower()
                if c == 'w':
                    total += current * long(604800)
                elif c == 'd':
                    total += current * long(86400)
                elif c == 'h':
                    total += current * long(3600)
                elif c == 'm':
                    total += current * long(60)
                elif c == 's':
                    total += current
                else:
                    raise BadTTL("unknown unit '%s'" % c)
                current = 0
        if not current == 0:
            raise BadTTL("trailing integer")
    if total < long(0) or total > long(2147483647):
        raise BadTTL("TTL should be between 0 and 2^31 - 1 (inclusive)")
    return total



class Payment:
    def __init__(
        self,
        payment_id,         # type: int
        name,               # type: str
        currency,           # type: str
        change_limit,       # type: float
        electronic_type,    # type: int
        open_drawer,        # type: bool
        need_auth,          # type: int
        parent_id,          # type: int
        custom_attr,        # type: str
    ):
        # type: (...) -> None

        self.payment_id = payment_id
        self.name = name
        self.currency = currency
        self.change_limit = change_limit
        self.electronic_type = electronic_type
        self.open_drawer = open_drawer
        self.need_auth = need_auth
        self.parent_id = parent_id
        self.custom_attr = custom_attr



# iban.py - functions for handling Norwegian IBANs
# coding: utf-8
#
# Copyright (C) 2018 Arthur de Jong
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

"""Norwegian IBAN (International Bank Account Number).

The IBAN is used to identify bank accounts across national borders. The
Norwegian IBAN is built up of the IBAN prefix (NO) and check digits, followed
by the 11 digit Konto nr. (bank account number).

>>> validate('NO93 8601 1117 947')
'NO9386011117947'
>>> to_kontonr('NO93 8601 1117 947')
'86011117947'
>>> format('NO9386011117947')
'NO93 8601 1117 947'
>>> validate('GR1601101050000010547023795')  # different country
Traceback (most recent call last):
    ...
InvalidComponent: ...
>>> validate('NO92 8601 1117 947')  # invalid IBAN check digit
Traceback (most recent call last):
    ...
InvalidChecksum: ...
>>> validate('NO23 8601 1117 946')  # invalid Konto nr. check digit
Traceback (most recent call last):
    ...
InvalidChecksum: ...
"""

from stdnum import iban
from stdnum.exceptions import *
from stdnum.no import kontonr


__all__ = ['compact', 'format', 'to_kontonr', 'validate', 'is_valid']


compact = iban.compact
format = iban.format


def to_kontonr(number):
    """Return the Norwegian bank account number part of the number."""
    number = compact(number)
    if not number.startswith('NO'):
        raise InvalidComponent()
    return number[4:]


def validate(number):
    """Check if the number provided is a valid Norwegian IBAN."""
    number = iban.validate(number, check_country=False)
    kontonr.validate(to_kontonr(number))
    return number


def is_valid(number):
    """Check if the number provided is a valid Norwegian IBAN."""
    try:
        return bool(validate(number))
    except ValidationError:
        return False



# cnic.py - functions for handling Pakistani CNIC numbers
# coding: utf-8
#
# Copyright (C) 2022 Syed Haseeb Shah
# Copyright (C) 2022 Arthur de Jong
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

"""CNIC number (Pakistani Computerised National Identity Card number).

The CNIC (Computerised National Identity Card, قومی شناختی کارڈ) or SNIC
(Smart National Identity Card) is issued by by Pakistan's NADRA (National
Database and Registration Authority) to citizens of 18 years or older.

The number consists of 13 digits and encodes the person's locality (5
digits), followed by 7 digit serial number an a single digit indicating
gender.

More Information:

* https://en.wikipedia.org/wiki/CNIC_(Pakistan)
* https://www.nadra.gov.pk/identity/identity-cnic/

>>> validate('34201-0891231-8')
'3420108912318'
>>> validate('42201-0397640-8')
'4220103976408'
>>> get_gender('42201-0397640-8')
'F'
>>> get_province('42201-0397640-8')
'Sindh'
>>> format('3420108912318')
'34201-0891231-8'
"""

from stdnum.exceptions import *
from stdnum.util import clean, isdigits


def compact(number):
    """Convert the number to the minimal representation. This strips the
    number of any valid separators and removes surrounding whitespace."""
    return clean(number, '-').strip()


def get_gender(number):
    """Get the person's birth gender ('M' or 'F')."""
    number = compact(number)
    if number[-1] in '13579':
        return 'M'
    elif number[-1] in '2468':
        return 'F'


# Valid Province IDs
PROVINCES = {
    '1': 'Khyber Pakhtunkhwa',
    '2': 'FATA',
    '3': 'Punjab',
    '4': 'Sindh',
    '5': 'Balochistan',
    '6': 'Islamabad',
    '7': 'Gilgit-Baltistan',
}


def get_province(number):
    """Get the person's birth gender ('M' or 'F')."""
    number = compact(number)
    return PROVINCES.get(number[0])


def validate(number):
    """Check if the number is a valid CNIC. This checks the length, formatting
    and some digits."""
    number = compact(number)
    if not isdigits(number):
        raise InvalidFormat()
    if len(number) != 13:
        raise InvalidLength()
    if not get_gender(number):
        raise InvalidComponent()
    if not get_province(number):
        raise InvalidComponent()
    return number


def is_valid(number):
    """Check if the number is a valid CNIC."""
    try:
        return bool(validate(number))
    except ValidationError:
        return False


def format(number):
    """Reformat the number to the standard presentation format."""
    number = compact(number)
    return '-'.join((number[:5], number[5:12], number[12:]))



import json
import inspect
from typing import (
    Optional,
    Union,
    List,
    Any,
)

from pyscripts import mbcontext
from posot import OrderTaker, OrderTakerException
from sysactions import send_message, get_podtype, get_operator_session, get_business_period
from systools import sys_log_info
from msgbus import TK_SYS_NAK, TK_CMP_MESSAGE, FM_STRING, FM_PARAM, TK_SYS_ACK, MBException

__all__ = ("TableService", "TableServiceException", "get_pos_ts")
DEFAULT_TIMEOUT = 300 * 1000 * 1000


class TableServiceException(OrderTakerException):
    def __init__(self, errmsg=""):
        super(TableServiceException, self).__init__("-1", errmsg)


class TableService(OrderTaker):
    def __init__(self, pos_id, mb_easy_context=None, order_mgr=None, product_mgr=None, table_mgr=None):
        super(TableService, self).__init__(pos_id, mb_easy_context, order_mgr, product_mgr)
        self.table_mgr = table_mgr or "TABLEMGR"

    def send_table_message(self, params):
        sys_log_info("TableService.{0}({1})".format(inspect.getframeinfo(inspect.currentframe()).function, params))
        req_buf = "\0".join([str(p) for p in params])
        msg = send_message(self.table_mgr, TK_CMP_MESSAGE, format=FM_PARAM, data=req_buf, timeout=DEFAULT_TIMEOUT)
        if msg.token == TK_SYS_NAK:
            message = "API exception: {0}({1})".format(inspect.getframeinfo(inspect.currentframe()).function, params)
            raise TableServiceException(errmsg=(msg.data if msg.format == FM_STRING and msg.data else message))

        return msg

    def listTables(self, user_id=None, table_id=None, with_props=None):
        params = ("LIST_TABLES", self._posid, user_id or "", table_id or "", with_props or "")
        msg = self.send_table_message(params)
        return json.loads(msg.data)

    def storePicture(self, force_update=False):
        params = ("STORE_PICTURE", self._posid, (1 if force_update else 0))
        msg = self.send_table_message(params)
        return msg.data

    def setCustomProperty(self, prop_key, prop_value="", table_id=None, service_id=None):
        params = ("SET_CUSTOM_PROPERTY", self._posid, prop_key, prop_value, table_id or "", service_id or "")
        self.send_table_message(params)

    def setCustomProperties(
        self,
        properties='',      # type: str
        table_id=None,      # type: str
        service_id=None,    # type: str
    ):
        # type: (...) -> None

        params = ("SET_CUSTOM_PROPERTIES", self._posid, properties, table_id or "", service_id or "")
        self.send_table_message(params)

    def getCustomProperty(self, prop_key, table_id=None, service_id=None):
        params = ("GET_CUSTOM_PROPERTY", self._posid, prop_key, table_id or "", service_id or "")
        msg = self.send_table_message(params)
        return json.loads(msg.data)

    def getTablePicture(self, table_id):
        params = ("TABLE_PICTURE", self._posid, table_id)
        msg = self.send_table_message(params)
        return msg.data

    def getServicePicture(self, service_id):
        params = ("SERVICE_PICTURE", self._posid, service_id)
        msg = self.send_table_message(params)
        return msg.data

    def getTableInfo(self, table_id):
        params = ("TABLE_INFO", self._posid, table_id)
        msg = self.send_table_message(params)
        return msg.data

    def recallService(self, user_id, table_id):
        params = ("RECALL_SERVICE", self._posid, user_id, table_id)
        msg = self.send_table_message(params)
        return msg.data

    def storeService(self, user_id, table_id, order_id=None):
        params = ("STORE_SERVICE", self._posid, user_id, table_id, order_id or "")
        self.send_table_message(params)
        return True

    def startService(self, user_id, b_period, table_id="", num_seats="", custom_props=None, source_table_id=""):
        props = [] if custom_props is None else json.dumps(custom_props)
        params = ("START_SERVICE", self._posid, user_id, table_id, b_period, num_seats, props, source_table_id)
        msg = self.send_table_message(params)
        return msg.data

    def sliceService(
        self,
        userid,                 # type: str
        new_setup,              # type: str
        src_table_id,           # type: str
        dst_table_id="",        # type: Optional[str]
        seat="",                # type: Optional[str]
        abandon_table=False,    # type: Optional[bool]
    ):
        params = (
            "SLICE_SERVICE",
            self._posid,
            userid,
            new_setup,
            src_table_id,
            dst_table_id,
            seat,
            abandon_table
        )
        msg = self.send_table_message(params=params)
        return json.loads(msg.data)

    def abandonService(self, user_id, table_id):
        params = ("ABANDON_SERVICE", self._posid, user_id, table_id)
        self.send_table_message(params)

    def createOrder(self, user_id, table_id, price_list=""):
        params = ("CREATE_ORDER", self._posid, user_id, table_id, price_list)
        msg = self.send_table_message(params)
        return msg.data

    def setParentTable(
        self,
        user_id,            # type: str
        table_id,           # type: str
        parent_table_id     # type: str
    ):
        # type: (...) -> None

        params = (
            "SET_PARENT_TABLE",
            self._posid,
            user_id,
            table_id,
            parent_table_id
        )
        self.send_table_message(params)

    def setCurrentOrder(self, user_id, table_id, order_id):
        params = ("SET_CURRENT_ORDER", self._posid, user_id, table_id, order_id)
        self.send_table_message(params)

    def voidServiceOrder(self, user_id, table_id, order_id):
        params = ("VOID_SERVICE_ORDER", self._posid, user_id, table_id, order_id)
        self.send_table_message(params)

    def registerServiceTender(self, user_id, table_id, tender_id, amount="", tender_detail=""):
        params = ("REGISTER_SERVICE_TENDER", self._posid, user_id, table_id, tender_id, amount, tender_detail)
        msg = self.send_table_message(params)
        return msg.data

    def updateTableTenders(self, user_id, table_id):
        params = ("UPDATE_TABLE_TENDERS", self._posid, user_id, table_id)
        msg = self.send_table_message(params)
        return msg.data

    def clearServiceTenders(self, user_id, table_id, service_tender_id="", delete_order_tender=True):
        params = ("CLEAR_SERVICE_TENDERS", self._posid, user_id, table_id, service_tender_id, delete_order_tender)
        msg = self.send_table_message(params)
        return msg.data

    def totalService(self, user_id, table_id, tip_rate="", tip_amount="", sale_type="EAT_IN"):
        params = ("TOTAL_SERVICE", self._posid, user_id, table_id, tip_rate, tip_amount, sale_type)
        msg = self.send_table_message(params)
        return msg.data

    def changeServiceTip(
        self,
        user_id,            # type: str
        table_id,           # type: str
        tip_rate="",        # type: str
        tip_amount="",      # type: str
        orders="",          # type: str
        manager_id="",      # type: str
    ):
        # type: (...) -> Any

        params = (
            "CHANGE_SERVICE_TIP",
            self._posid,
            user_id,
            table_id,
            tip_rate or "",
            tip_amount or "",
            orders or "",
            manager_id or "",
        )

        msg = self.send_table_message(
            params=params,
        )

        return msg.data

    def closeService(self, user_id, table_id):
        params = ("CLOSE_SERVICE", self._posid, user_id, table_id)
        msg = self.send_table_message(params)
        return msg.data

    def reopenService(self, user_id, table_id):
        params = ("REOPEN_SERVICE", self._posid, user_id, table_id)
        msg = self.send_table_message(params)
        return msg.data

    def joinTables(self, user_id, from_table_id, dest_table_id, number_of_seats=""):
        params = ("JOIN_TABLES", self._posid, user_id, from_table_id, dest_table_id, number_of_seats)
        self.send_table_message(params)

    def moveTable(
        self,
        user_id,                        # type: int
        from_table_id,                  # type: str
        dest_table_id,                  # type: str
        number_of_seats="",             # type: Optional[str]
        tables_to_set_available=None,   # type: Optional[List]
        mode="seats",                   # type: Optional[str]
        ignore_totaled_orders=False     # type: Optional[bool]
    ):
        params = (
            "MOVE_TABLE",
            self._posid,
            user_id,
            from_table_id,
            dest_table_id,
            number_of_seats,
            tables_to_set_available,
            mode,
            ignore_totaled_orders
        )
        self.send_table_message(params)

    def moveTableOrder(
            self,
            user_id,            # type: int
            from_table_id,      # type: str
            dest_table_id,      # type: str
            order_id,           # type: int
            number_of_seats=""  # type: Optional[Union[str, int]]
    ):
        # type: (...) -> None

        params = ("MOVE_TABLE_ORDER", self._posid, user_id, from_table_id, dest_table_id, order_id, number_of_seats)
        self.send_table_message(params)

    def moveChildTable(
            self,
            user_id,            # type: int
            from_table_id,      # type: str
            dest_table_id,      # type: str
            child_table_id,     # type: int
            custom_props=None   # type: str
    ):
        # type: (...) -> None

        props = [] if custom_props is None else json.dumps(custom_props)
        params = ("MOVE_CHILD_TABLE", self._posid, user_id, from_table_id, dest_table_id, child_table_id, props)
        self.send_table_message(params)

    def unlinkTables(self, user_id, table_id, unlink_table_id=""):
        params = ("UNLINK_TABLES", self._posid, user_id, table_id, unlink_table_id)
        self.send_table_message(params)

    def tableReady4Service(self, user_id, table_id):
        params = ("TABLE_READY4SERVICE", self._posid, user_id, table_id)
        self.send_table_message(params)

    def changeTableUser(self, user_id, table_id="", service_id=""):
        params = ("CHANGE_TABLE_USER", self._posid, user_id, table_id, service_id)
        self.send_table_message(params)

    def changeNumberOfSeats(self, user_id, number_of_seats, table_id="", service_id=""):
        params = ("CHANGE_NUMBER_OF_SEATS", self._posid, user_id, number_of_seats, table_id, service_id)
        self.send_table_message(params)

    def moveOrderToService(self, user_id, order_id, table_id="", service_id=""):
        params = ("MOVE_ORDER_TO_SERVICE", self._posid, user_id, table_id, order_id, service_id)
        self.send_table_message(params)

    def getSectors(
        self,
    ):
        # type: (...) -> List[str]

        params = ("GET_SECTORS", self._posid)
        msg = self.send_table_message(params=params)
        return msg.data

    def getOperatorSectorInfo(
        self,
        force=False,    # type: Optional[bool]
    ):
        # type: (...) -> List[str]

        params = ("GET_OPERATOR_SECTOR_INFO", self._posid, force)
        msg = self.send_table_message(params=params)
        return msg.data

    def updateOperatorSectorInfo(
        self,
        data,   # type: str
    ):
        # type: (...) -> List[str]

        params = ("UPDATE_OPERATOR_SECTOR_INFO", self._posid, data)
        msg = self.send_table_message(params=params)
        return msg.data

    def reset_store_layout(
        self,
    ):
        # type: (...) -> None

        params = ("RESET_STORE_LAYOUT", self._posid)
        self.send_table_message(params=params)

def get_pos_ts(model, order_mgr=None, product_mgr=None, table_mgr=None):
    pos_id = model.get("posId", "1") if model is not None else "1"
    pos_ot = TableService(pos_id, mbcontext, order_mgr, product_mgr, table_mgr)
    if model is not None:
        pos_ot.podType = get_podtype(model)
        pos_ot.businessPeriod = get_business_period(model)
        pos_ot.sessionId = get_operator_session(model)
        operator = model.find("Operator")
        if operator is not None:
            pos_ot.operatorId = operator.get("id")

    return pos_ot



# -*- coding: utf-8 -*-
import json
import logging
from typing import Optional

from sysactions import (
    action,
    show_any_dialog,
    show_messagebox,
    StopAction,
)

from actions.delivery.delivery_address_information import (
    insert_customer_and_save_on_sale,
    has_delivery_info,
)
from custom_sysactions import block_action_if_order_items_not_resolved

logger = logging.getLogger("PosActions")


@action
@block_action_if_order_items_not_resolved
def delivery_address(
    pos_id,             # type: str
    is_edit="false",    # type: Optional[str]
):
    # type: (...) -> bool

    from posactions import ask_customer_delivery_data

    is_edit = is_edit.lower() == "true"
    if is_edit:
        return request_and_save_delivery_address(pos_id=pos_id)

    if ask_customer_delivery_data == "onTotal" and not has_delivery_info(pos_id=pos_id):
        return request_and_save_delivery_address(pos_id=pos_id)
    return True


@action
def save_delivery_address(
    pos_id,                      # type: str
    is_edit="false",             # type: Optional[str]
    is_action="false",           # type: Optional[str]
    create_new_order="false",    # type: Optional[str]
):
    # type: (...) -> bool

    from posactions import ask_customer_delivery_data

    is_edit = is_edit.lower() == "true"
    is_action = is_action.lower() == "true"
    create_new_order = create_new_order.lower() == "true"

    if ask_customer_delivery_data == "onTotal" or is_action:
        return request_and_save_delivery_address(
            pos_id=pos_id,
            dialog_type="saveCustomerDeliveryAddress",
            create_new_order=create_new_order,
        )
    if is_edit:
        return request_and_save_delivery_address(pos_id=pos_id)
    return True


def request_and_save_delivery_address(
    pos_id,                           # type: str
    dialog_type="deliveryAddress",    # type: Optional[str]
    create_new_order=False,           # type: Optional[bool]
):
    # type: (...) -> bool

    try:
        customer = show_any_dialog(
            posid=pos_id,
            dlgtype=dialog_type,
            message="",
            title="",
            infotext="",
            buttons="",
            icon="",
            timeout=6000000,
            mask="",
            defvalue="",
            minvalue="",
            maxvalue="",
            contents="",
            asynch=False,
        )
        customer_json = json.loads(customer)

        if customer_json:
            selected_option = customer_json.get('SELECTED_OPTION', None)
            if selected_option is None:
                raise StopAction()

            insert_customer_and_save_on_sale(
                pos_id=pos_id,
                customer_json=customer_json,
                create_new_order=create_new_order,
            )

        return True
    except Exception as _:
        logger.exception("Error inserting or updating customer")
        show_messagebox(pos_id, message="$CUSTOMER_ERROR_ON_UPDATE")
        return False



from msgbus import (
    TK_SYS_NAK,
    TK_CMP_MESSAGE,
    FM_PARAM,
)
from sysactions import send_message


def get_table_picture(
    pos_id,    # type: str
    table_id,  # type: str
):
    # type: (...) -> str

    params = ["TABLE_PICTURE", pos_id, table_id]
    req_buf = "\0".join([str(p) for p in params])
    result = send_message(
        dest_name="TABLEMGR",
        token=TK_CMP_MESSAGE,
        format=FM_PARAM,
        data=req_buf,
    )
    if result.token == TK_SYS_NAK:
        message = "Error getting table_picture"
        raise Exception(message)

    return result.data



# cr.py - functions for handling Costa Rica DIMEX numbers
# coding: utf-8
#
# Copyright (C) 2019 Leandro Regueiro
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

"""CR (Cédula de Residencia, Costa Rica foreigners ID number).

The Cédula de Residencia (CR), also know as DIMEX or Documento de
Identificación Migratorio para Extranjeros, is an identifier of foreigners in
Costa Rica.

This number consists of 11 or 12 digits in the form 1NNN-CC...C-EE...E where
NNN represents the code of the country the foreigner comes from as specified
by Costa Rica's Dirección General de Migración y Extranjería, CC...C is a
sequence telling how many Cédula de Residencia have been issued in total and
EE...E is a sequence telling how many Cédula de Residencia have been issued
for that particular foreign country.

More information:

* https://www.hacienda.go.cr/consultapagos/ayuda_cedulas.htm
* https://www.procomer.com/downloads/quiero/guia_solicitud_vuce.pdf (page 12)
* https://www.hacienda.go.cr/ATV/frmConsultaSituTributaria.aspx

>>> validate('155812994816')
'155812994816'
>>> validate('30123456789')
Traceback (most recent call last):
    ...
InvalidComponent: ...
>>> validate('12345678')
Traceback (most recent call last):
    ...
InvalidLength: ...
>>> format('122200569906')
'122200569906'
"""

from stdnum.exceptions import *
from stdnum.util import clean, isdigits


def compact(number):
    """Convert the number to the minimal representation.

    This strips the number of any valid separators and removes surrounding
    whitespace.
    """
    return clean(number, ' -').upper().strip()


def validate(number):
    """Check if the number is a valid Costa Rica CR number.

    This checks the length and formatting.
    """
    number = compact(number)
    if len(number) not in (11, 12):
        raise InvalidLength()
    if not isdigits(number):
        raise InvalidFormat()
    if number[0] != '1':
        raise InvalidComponent()
    return number


def is_valid(number):
    """Check if the number is a valid Costa Rica CR number."""
    try:
        return bool(validate(number))
    except ValidationError:
        return False


def format(number):
    """Reformat the number to the standard presentation format."""
    return compact(number)
